# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Trestle Create CommandPlusDocs."""

import argparse
import logging
from datetime import datetime
from typing import Type

import trestle.common.err as err
import trestle.common.file_utils
import trestle.oscal
from trestle.common import const, file_utils, log
from trestle.common.common_types import TopLevelOscalModel
from trestle.common.model_utils import ModelUtils
from trestle.core import generators
from trestle.core.commands.add import Add
from trestle.core.commands.command_docs import CommandPlusDocs
from trestle.core.commands.common.return_codes import CmdReturnCodes
from trestle.core.models.actions import CreatePathAction, WriteFileAction
from trestle.core.models.elements import Element, ElementPath
from trestle.core.models.file_content_type import FileContentType
from trestle.core.models.plans import Plan

logger = logging.getLogger(__name__)


class CreateCmd(CommandPlusDocs):
    """Create a sample OSCAL model in trestle project or create new elements within a given model."""

    name = 'create'

    def _init_arguments(self) -> None:
        self.add_argument('-t', '--type', help='Type of model if created anew.', choices=const.MODEL_TYPE_LIST)
        self.add_argument('-o', '--output', help='Name of the output created model.')
        self.add_argument(const.IOF_SHORT, const.IOF_LONG, help=const.IOF_HELP, action='store_true')
        self.add_argument(
            '-x', '--extension', help='Type of file output.', choices=['json', 'yaml', 'yml'], default='json'
        )
        self.add_argument(
            '-f', '--file', help='Optional existing OSCAL file that will have elements created within it.', type=str
        )
        self.add_argument(
            '-e', '--element', help='Optional path of element to be created whithin the specified file.', type=str
        )

    def _run(self, args: argparse.Namespace) -> int:
        """
        Execute the create command.

        Notes
            Either a new model will be created of the specified type,
            or an existing file will have new elements added within it.
        """
        try:
            # Normal create path
            if args.type and args.output:
                object_type = ElementPath(args.type).get_type()
                return self.create_object(args.type, object_type, args)
            # Add path
            elif args.file and args.element:
                add = Add()
                return add.add_from_args(args)

            raise err.TrestleIncorrectArgsError(
                'Create requires either a model type and output name, or a file and element path.'
            )

        except Exception as e:  # pragma: no cover
            return err.handle_generic_command_exception(e, logger, 'Error while creating a sample OSCAL model')

    @classmethod
    def create_object(cls, model_alias: str, object_type: Type[TopLevelOscalModel], args: argparse.Namespace) -> int:
        """Create a top level OSCAL object within the trestle directory, leveraging functionality in add."""
        log.set_log_level_from_args(args)
        trestle_root = args.trestle_root  # trestle root is set via command line in args. Default is cwd.
        if not trestle_root or not file_utils.is_valid_project_root(args.trestle_root):
            raise err.TrestleRootError(f'Given directory {trestle_root} is not a trestle project.')

        plural_path = ModelUtils.model_type_to_model_dir(model_alias)

        desired_model_dir = trestle_root / plural_path / args.output

        desired_model_path = desired_model_dir / (model_alias + '.' + args.extension)

        if desired_model_path.exists():
            raise err.TrestleError(f'OSCAL file to be created here: {desired_model_path} exists.')

        # Create sample model.
        sample_model = generators.generate_sample_model(object_type, include_optional=args.include_optional_fields)
        # Presuming top level level model not sure how to do the typing for this.
        sample_model.metadata.title = f'Generic {model_alias} created by trestle named {args.output}.'  # type: ignore
        sample_model.metadata.last_modified = datetime.now().astimezone()
        sample_model.metadata.oscal_version = trestle.oscal.OSCAL_VERSION
        sample_model.metadata.version = '0.0.0'

        top_element = Element(sample_model, model_alias)

        create_action = CreatePathAction(desired_model_path.resolve(), True)
        write_action = WriteFileAction(
            desired_model_path.resolve(), top_element, FileContentType.to_content_type(desired_model_path.suffix)
        )

        # create a plan to write the directory and file.
        create_plan = Plan()
        create_plan.add_action(create_action)
        create_plan.add_action(write_action)
        create_plan.execute()
        return CmdReturnCodes.SUCCESS.value



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for trestle author docs subcommand."""
import pathlib
import shutil
import sys
from uuid import uuid4

from _pytest.monkeypatch import MonkeyPatch

import frontmatter

import pytest

from tests.test_utils import execute_command_and_assert

import trestle
from trestle.core.commands.author.consts import START_TEMPLATE_VERSION
from trestle.core.markdown.markdown_api import MarkdownAPI

import yaml


@pytest.mark.parametrize(
    'command_string, return_code',
    [
        ('trestle author docs setup', 2), ('trestle author docs setup --task-name foobaa', 0),
        ('trestle author docs setup --task-name catalogs', 1)
    ]
)
def test_governed_docs_high(
    tmp_trestle_dir: pathlib.Path, command_string: str, return_code: int, monkeypatch: MonkeyPatch
) -> None:
    """Simple execution tests of trestle author docs."""
    monkeypatch.setattr(sys, 'argv', command_string.split())
    with pytest.raises(SystemExit) as wrapped_error:
        trestle.cli.run()
    assert wrapped_error.type == SystemExit
    assert wrapped_error.value.code == return_code


@pytest.mark.parametrize(
    'task_name, template_content, target_content, recurse, setup_code, template_code, validate_code',
    [
        (
            'test_task',
            pathlib.Path('author/0.0.1/test_1_md_format/template.md'),
            pathlib.Path('author/0.0.1/test_1_md_format/correct_instance_extra_features.md'),
            False,
            0,
            0,
            0
        ),
        (
            'catalogs',
            pathlib.Path('author/0.0.1/test_1_md_format/template.md'),
            pathlib.Path('author/0.0.1/test_1_md_format/correct_instance_extra_features.md'),
            False,
            1,
            1,
            1,
        ),
        (
            'test_task',
            pathlib.Path('author/0.0.1/test_1_md_format/template.md'),
            pathlib.Path('author/0.0.1/test_1_md_format/bad_instance_missing_heading.md'),
            False,
            0,
            0,
            1
        ),
        (
            'test_task',
            pathlib.Path('author/utf16test/sample_okay.md'),
            pathlib.Path('author/utf16test/sample_utf16.md'),
            False,
            0,
            0,
            1
        ),
        (
            'test_task',
            pathlib.Path('author/utf16test/sample_utf16.md'),
            pathlib.Path('author/utf16test/sample_okay.md'),
            False,
            0,
            1,
            1
        ),
        (
            'test_task',
            pathlib.Path('author/0.0.1/test_1_md_format/template.md'),
            pathlib.Path('author/0.0.1/test_1_md_format/correct_instance_extra_features.md'),
            True,
            0,
            0,
            0
        ),
        (
            'test_task',
            pathlib.Path('author/0.0.1/test_1_md_format/template.md'),
            pathlib.Path('author/0.0.1/test_1_md_format/correct_instance_extra_features.md'),
            True,
            0,
            0,
            1
        )
    ]
)
def test_e2e(
    task_name: str,
    template_content: pathlib.Path,
    target_content: pathlib.Path,
    recurse: bool,
    setup_code: int,
    template_code: int,
    validate_code: int,
    testdata_dir: pathlib.Path,
    tmp_trestle_dir: pathlib.Path,
    monkeypatch: MonkeyPatch
) -> None:
    """Run an E2E workflow with two test criteria for success."""
    # Note testdata_dir must be before tmp_trestle_dir in the argument order.
    recurse_flag = '-r' if recurse else ''
    command_string_setup = f'trestle author docs setup -tn {task_name}'
    command_string_create_sample = f'trestle author docs create-sample -tn {task_name}'
    command_string_validate_template = f'trestle author docs template-validate -tn {task_name}'
    command_string_validate_content = (f'trestle author docs validate -tn {task_name} --header-validate {recurse_flag}')
    template_target_loc = tmp_trestle_dir / '.trestle' / 'author' / task_name / START_TEMPLATE_VERSION / 'template.md'
    test_content_loc = tmp_trestle_dir / task_name / f'{uuid4()}.md'
    # Test setup
    monkeypatch.setattr(sys, 'argv', command_string_setup.split())
    with pytest.raises(SystemExit) as wrapped_error:
        trestle.cli.run()
    assert wrapped_error.type == SystemExit
    assert wrapped_error.value.code == setup_code
    if setup_code > 0:
        return
    # Copy in template:
    shutil.copyfile(str(testdata_dir / template_content), str(template_target_loc))
    monkeypatch.setattr(sys, 'argv', command_string_validate_template.split())
    rc = trestle.cli.Trestle().run()
    assert rc == template_code
    if template_code > 0:
        return

    # Create sample - should always work if we are here.
    monkeypatch.setattr(sys, 'argv', command_string_create_sample.split())
    rc = trestle.cli.Trestle().run()
    assert rc == 0

    shutil.copyfile(str(testdata_dir / target_content), str(test_content_loc))
    if recurse:
        # choose source file based on expected result
        sub_file_name = 'bad_instance_missing_heading.md' if validate_code else 'correct_instance_extra_features.md'
        sub_file_path = testdata_dir / 'author/0.0.1/test_1_md_format' / sub_file_name
        subdir = tmp_trestle_dir / task_name / 'subdir'
        subdir.mkdir()
        sub_content = subdir / f'{uuid4()}.md'
        shutil.copyfile(str(sub_file_path), str(sub_content))

    monkeypatch.setattr(sys, 'argv', command_string_validate_content.split())
    rc = trestle.cli.Trestle().run()
    assert rc == validate_code


def test_failure_bad_template_dir(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Create and test a bad directory."""
    setup_ = 'trestle author docs setup --task-name foobaa'

    bad_template_sub_directory = tmp_trestle_dir / '.trestle' / 'author' / 'foobaa' / 'test'
    bad_template_sub_directory.mkdir(parents=True, exist_ok=True)
    monkeypatch.setattr(sys, 'argv', setup_.split())
    with pytest.raises(SystemExit) as wrapped_error:
        trestle.cli.run()
    assert wrapped_error.type == SystemExit
    assert wrapped_error.value.code == 1


def test_success_repeated_call(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test double setup."""
    setup_ = 'trestle author docs setup --task-name foobaa'

    monkeypatch.setattr(sys, 'argv', setup_.split())
    with pytest.raises(SystemExit) as wrapped_error:
        trestle.cli.run()
    assert wrapped_error.type == SystemExit
    assert wrapped_error.value.code == 0
    monkeypatch.setattr(sys, 'argv', setup_.split())
    with pytest.raises(SystemExit) as wrapped_error:
        trestle.cli.run()
    assert wrapped_error.type == SystemExit
    assert wrapped_error.value.code == 0


def test_e2e_debugging(testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Run an E2E workflow with two test criteria for success."""
    # hardcoded arguments for testing
    task_name = 'test_task'
    recurse = False
    template_content = pathlib.Path('author/0.0.1/test_1_md_format/template.md')
    target_content = pathlib.Path('author/0.0.1/test_1_md_format/correct_instance_extra_features.md')
    setup_code = 0
    template_code = 0
    validate_code = 0

    # Note testdata_dir must be before tmp_trestle_dir in the argument order.
    recurse_flag = '-r' if recurse else ''
    command_string_setup = f'trestle author docs setup -tn {task_name}'
    command_string_create_sample = f'trestle author docs create-sample -tn {task_name}'
    command_string_validate_template = f'trestle author docs template-validate -tn {task_name}'
    command_string_validate_content = (f'trestle author docs validate -tn {task_name} --header-validate {recurse_flag}')
    template_target_loc = tmp_trestle_dir / '.trestle' / 'author' / task_name / 'template.md'
    test_content_loc = tmp_trestle_dir / task_name / f'{uuid4()}.md'
    # Test setup
    monkeypatch.setattr(sys, 'argv', command_string_setup.split())
    with pytest.raises(SystemExit) as wrapped_error:
        trestle.cli.run()
    assert wrapped_error.type == SystemExit
    assert wrapped_error.value.code == setup_code
    if setup_code > 0:
        return
    # Copy in template:
    shutil.copyfile(str(testdata_dir / template_content), str(template_target_loc))

    monkeypatch.setattr(sys, 'argv', command_string_validate_template.split())
    with pytest.raises(SystemExit) as wrapped_error:
        trestle.cli.run()
    assert wrapped_error.type == SystemExit
    assert wrapped_error.value.code == template_code
    if template_code > 0:
        return
    # Create sample - should always work if we are here.
    monkeypatch.setattr(sys, 'argv', command_string_create_sample.split())
    with pytest.raises(SystemExit) as wrapped_error:
        trestle.cli.run()
    assert wrapped_error.type == SystemExit
    assert wrapped_error.value.code == 0

    shutil.copyfile(str(testdata_dir / target_content), str(test_content_loc))
    if recurse:
        # choose source file based on expected result
        sub_file_name = 'bad_instance_missing_heading.md' if validate_code else 'correct_instance_extra_features.md'
        sub_file_path = testdata_dir / 'author/0.0.1/test_1_md_format' / sub_file_name
        subdir = tmp_trestle_dir / task_name / 'subdir'
        subdir.mkdir()
        sub_content = subdir / f'{uuid4()}.md'
        shutil.copyfile(str(sub_file_path), str(sub_content))
    monkeypatch.setattr(sys, 'argv', command_string_validate_content.split())
    with pytest.raises(SystemExit) as wrapped_error:
        trestle.cli.run()
    assert wrapped_error.type == SystemExit
    assert wrapped_error.value.code == validate_code


@pytest.mark.parametrize(
    'task_name, template_source_directory, target_source_directory, readme_validate, template_code, validate_code',
    [
        (
            'test_task_0',
            pathlib.Path('author/docs/template_folder_valid_with_readme'),
            pathlib.Path('author/docs/candidate_valid_readme'),
            False,
            0,
            0
        ),
        (
            'test_task_1',
            pathlib.Path('author/docs/template_folder_valid_with_readme'),
            pathlib.Path('author/docs/candidate_valid_readme'),
            True,
            0,
            1
        ),
        (
            'test_task_2',
            pathlib.Path('author/docs/template_folder_invalid'),
            pathlib.Path('author/docs/candidate_valid_readme'),
            False,
            1,
            1
        ),
        (
            'test_task_3',
            pathlib.Path('author/docs/template_folder_invalid'),
            pathlib.Path('author/docs/candidate_valid_readme'),
            True,
            1,
            1
        ),
        (
            'test_task_4',
            pathlib.Path('author/docs/template_folder_valid_with_readme'),
            pathlib.Path('author/docs/candidate_invalid_readme'),
            False,
            0,
            1
        ),
        (
            'test_task_5',
            pathlib.Path('author/docs/template_folder_valid_with_readme'),
            pathlib.Path('author/docs/candidate_invalid_readme'),
            True,
            0,
            1
        )
    ]
)
def test_directory_content(
    task_name: str,
    template_source_directory: pathlib.Path,
    target_source_directory: pathlib.Path,
    readme_validate: bool,
    template_code: int,
    validate_code: int,
    testdata_dir: pathlib.Path,
    tmp_trestle_dir: pathlib.Path,
    monkeypatch: MonkeyPatch
):
    """Test setup where groups of files are tested to allow testing of readme flags."""
    readme_validate_flag = '-rv' if readme_validate else ''
    command_string_create_sample = f'trestle author docs create-sample -tn {task_name}'
    command_string_validate_template = f'trestle author docs template-validate -tn {task_name}'
    command_string_validate_content = (
        f'trestle author docs validate -tn {task_name} --header-validate {readme_validate_flag}'
    )
    template_target_directory = tmp_trestle_dir / '.trestle' / 'author' / task_name
    task_directory = tmp_trestle_dir / task_name
    # Test setup
    shutil.copytree(str(testdata_dir / template_source_directory), str(template_target_directory))
    shutil.copytree(str(testdata_dir / target_source_directory), str(task_directory))
    monkeypatch.setattr(sys, 'argv', command_string_validate_template.split())
    with pytest.raises(SystemExit) as wrapped_error:
        trestle.cli.run()
    assert wrapped_error.type == SystemExit
    assert wrapped_error.value.code == template_code
    if template_code > 0:
        return
    # Create sample - should always work if we are here.
    monkeypatch.setattr(sys, 'argv', command_string_create_sample.split())
    with pytest.raises(SystemExit) as wrapped_error:
        trestle.cli.run()
    assert wrapped_error.type == SystemExit
    assert wrapped_error.value.code == 0

    monkeypatch.setattr(sys, 'argv', command_string_validate_content.split())
    with pytest.raises(SystemExit) as wrapped_error:
        trestle.cli.run()
    assert wrapped_error.type == SystemExit
    assert wrapped_error.value.code == validate_code


@pytest.mark.parametrize(
    'task_name, template_content, target_content, recurse, setup_code, template_code, validate_code',
    [
        (
            'test_task',
            pathlib.Path('author/0.0.1/test_1_md_format/template.md'),
            pathlib.Path('author/0.0.1/test_1_md_format/correct_instance_extra_features.md'),
            False,
            0,
            0,
            0
        ),
        (
            'catalogs',
            pathlib.Path('author/0.0.1/test_1_md_format/template.md'),
            pathlib.Path('author/0.0.1/test_1_md_format/correct_instance_extra_features.md'),
            False,
            1,
            1,
            1,
        ),
        (
            'test_task',
            pathlib.Path('author/0.0.1/test_1_md_format/template.md'),
            pathlib.Path('author/0.0.1/test_1_md_format/bad_instance_missing_heading.md'),
            False,
            0,
            0,
            1
        ),
        (
            'test_task',
            pathlib.Path('author/utf16test/sample_okay.md'),
            pathlib.Path('author/utf16test/sample_utf16.md'),
            False,
            0,
            0,
            1
        ),
        (
            'test_task',
            pathlib.Path('author/utf16test/sample_utf16.md'),
            pathlib.Path('author/utf16test/sample_okay.md'),
            False,
            0,
            1,
            1
        ),
        (
            'test_task',
            pathlib.Path('author/0.0.1/test_1_md_format/template.md'),
            pathlib.Path('author/0.0.1/test_1_md_format/correct_instance_extra_features.md'),
            True,
            0,
            0,
            0
        ),
        (
            'test_task',
            pathlib.Path('author/0.0.1/test_1_md_format/template.md'),
            pathlib.Path('author/0.0.1/test_1_md_format/correct_instance_extra_features.md'),
            True,
            0,
            0,
            1
        )
    ]
)
def test_e2e_backward_compatibility(
    task_name: str,
    template_content: pathlib.Path,
    target_content: pathlib.Path,
    recurse: bool,
    setup_code: int,
    template_code: int,
    validate_code: int,
    testdata_dir: pathlib.Path,
    tmp_trestle_dir: pathlib.Path,
    monkeypatch: MonkeyPatch
) -> None:
    """Same as E2E test but pretend like workspace existed and needs to be updated."""
    # Note testdata_dir must be before tmp_trestle_dir in the argument order.
    recurse_flag = '-r' if recurse else ''
    command_string_setup = f'trestle author docs setup -tn {task_name}'
    command_string_create_sample = f'trestle author docs create-sample -tn {task_name}'
    command_string_validate_template = f'trestle author docs template-validate -tn {task_name}'
    command_string_validate_content = (f'trestle author docs validate -tn {task_name} --header-validate {recurse_flag}')
    template_target_loc = tmp_trestle_dir / '.trestle' / 'author' / task_name / 'template.md'
    test_content_loc = tmp_trestle_dir / task_name / f'{uuid4()}.md'

    # Copy in template using old path
    (tmp_trestle_dir / '.trestle' / 'author' / task_name).mkdir(exist_ok=True, parents=True)
    shutil.copyfile(str(testdata_dir / template_content), str(template_target_loc))

    # Test setup
    monkeypatch.setattr(sys, 'argv', command_string_setup.split())
    rc = trestle.cli.Trestle().run()
    assert rc == setup_code
    if setup_code > 0:
        return
    assert not template_target_loc.exists()
    assert (tmp_trestle_dir / '.trestle' / 'author' / task_name / START_TEMPLATE_VERSION / 'template.md').exists()

    execute_command_and_assert(command_string_validate_template, template_code, monkeypatch)
    if template_code > 0:
        return

    # Create sample - should always work if we are here.
    execute_command_and_assert(command_string_create_sample, 0, monkeypatch)

    shutil.copyfile(str(testdata_dir / target_content), str(test_content_loc))
    if recurse:
        # choose source file based on expected result
        sub_file_name = 'bad_instance_missing_heading.md' if validate_code else 'correct_instance_extra_features.md'
        sub_file_path = testdata_dir / 'author/0.0.1/test_1_md_format' / sub_file_name
        subdir = tmp_trestle_dir / task_name / 'subdir'
        subdir.mkdir()
        sub_content = subdir / f'{uuid4()}.md'
        shutil.copyfile(str(sub_file_path), str(sub_content))

    execute_command_and_assert(command_string_validate_content, validate_code, monkeypatch)


def test_ignore_flag(testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test that ignored files are not validated. Validation will fail if attempted."""
    task_template_folder = tmp_trestle_dir / '.trestle/author/test_task/'
    author_template = task_template_folder / 'template.md'
    test_data_folder = testdata_dir / 'author/0.0.1/test_5_md_ignored'
    task_instance_folder = tmp_trestle_dir / 'test_task'
    task_template_folder.mkdir(exist_ok=True, parents=True)
    author_template.touch()
    shutil.copyfile(test_data_folder / 'template.md', author_template)

    # copy all files
    shutil.copytree(test_data_folder, task_instance_folder)

    command_string_validate_content = 'trestle author docs validate -tn test_task -hv -ig ^_.*'
    execute_command_and_assert(command_string_validate_content, 0, monkeypatch)

    command_string_validate_content = 'trestle author docs validate -tn test_task -hv -ig ^_.* -r'
    execute_command_and_assert(command_string_validate_content, 0, monkeypatch)

    command_string_validate_content = 'trestle author docs validate -tn test_task -hv -r'
    execute_command_and_assert(command_string_validate_content, 1, monkeypatch)


def test_fail_when_no_task(testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test that validation fails when task does not exists."""
    command_validate = 'trestle author docs validate -tn test_task'
    execute_command_and_assert(command_validate, 1, monkeypatch)


def test_fail_when_no_template(
    testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """Test that validation fails when template does not exists."""
    task_template_folder = tmp_trestle_dir / '.trestle/author/test_task/'
    source_instance = testdata_dir / 'author/0.0.1/test_1_md_format/correct_instance.md'
    task_instance = tmp_trestle_dir / 'test_task/correct_instance.md'
    task_template_folder.mkdir(exist_ok=True, parents=True)
    (tmp_trestle_dir / 'test_task').mkdir(parents=True)
    task_instance.touch()
    shutil.copyfile(source_instance, task_instance)

    command_validate = 'trestle author docs validate -tn test_task'
    execute_command_and_assert(command_validate, 1, monkeypatch)


def test_instance_no_header(
    testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """Test behaviour when validating instance with no header."""
    test_data_folder = testdata_dir / 'author/0.0.1/test_1_md_format'
    bad_source_instance = test_data_folder / 'bad_instance_no_header.md'

    task_template_folder = tmp_trestle_dir / '.trestle/author/test_task/'
    task_instance_folder = tmp_trestle_dir / 'test_task'
    task_template = task_template_folder / 'template.md'
    task_instance = task_instance_folder / 'no_header_instance.md'
    task_template_folder.mkdir(exist_ok=True, parents=True)
    task_instance_folder.mkdir(parents=True)

    shutil.copy(test_data_folder / 'template.md', task_template)
    shutil.copy(bad_source_instance, task_instance)

    command_string_validate_content = 'trestle author docs validate -tn test_task -hv'
    execute_command_and_assert(command_string_validate_content, 1, monkeypatch)

    command_string_validate_content = 'trestle author docs validate -tn test_task'
    execute_command_and_assert(command_string_validate_content, 0, monkeypatch)

    command_string_validate_content = 'trestle author docs validate -tn test_task -tv 0.0.1'
    execute_command_and_assert(command_string_validate_content, 0, monkeypatch)

    command_string_validate_content = 'trestle author docs validate -tn test_task -tv 0.0.1 -hv'
    execute_command_and_assert(command_string_validate_content, 1, monkeypatch)

    # place bad instance in subfolder
    good_source_instance = test_data_folder / 'correct_instance.md'
    shutil.copyfile(good_source_instance, task_instance)
    (task_instance_folder / 'subfolder').mkdir(parents=True)
    shutil.copy(bad_source_instance, task_instance_folder / 'subfolder/bad_instance.md')

    command_string_validate_content = 'trestle author docs validate -tn test_task -hv -r'
    execute_command_and_assert(command_string_validate_content, 1, monkeypatch)


def test_template_validate_no_template(
    testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """Test that template validation fails when no template."""
    task_template_folder = tmp_trestle_dir / '.trestle/author/test_task/'
    task_template_folder.mkdir(exist_ok=True, parents=True)
    command_validate = 'trestle author docs template-validate -tn test_task'
    execute_command_and_assert(command_validate, 1, monkeypatch)


def test_001_template_folder_is_not_created(
    testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """Test that folder 0.0.1 is not recreated."""
    command = 'trestle author docs setup -tn my_task'
    execute_command_and_assert(command, 0, monkeypatch)

    # move template
    template_folder_old = tmp_trestle_dir.joinpath('.trestle/author/my_task/0.0.1')
    template_folder_new = tmp_trestle_dir.joinpath('.trestle/author/my_task/0.1.0')
    template_folder_new.mkdir(parents=True)
    template_path_old = template_folder_old.joinpath('template.md')
    template_path_new = template_folder_new.joinpath('template.md')

    shutil.copy(template_path_old, template_folder_new)

    # remove old template folder
    shutil.rmtree(template_folder_old)

    # modify the header in the template
    contents = frontmatter.loads(template_path_new.open('r').read())
    header = contents.metadata
    header['x-trestle-template-version'] = '0.1.0'

    # write file back
    with open(template_path_new, 'w') as template:
        template.write('---\n')
        yaml.dump(header, template, sort_keys=False)
        template.write('---\n\n')
        template.write(contents.content)

    command = 'trestle author docs validate -tn my_task'
    execute_command_and_assert(command, 0, monkeypatch)

    # ensure that 0.0.1 is not recreated and content is preserved
    assert not template_folder_old.exists()
    assert template_folder_new.exists()
    assert template_folder_new.joinpath('template.md').exists()

    contents = frontmatter.loads(template_path_new.open('r').read())
    header = contents.metadata
    assert header['x-trestle-template-version'] == '0.1.0'


def test_instance_validate_with_governed_heading(
    testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch
) -> None:
    """Test behaviour when validating good instance with governed heading and various changes."""
    task_template = tmp_trestle_dir / '.trestle/author/test_task/template.md'
    test_template_folder = testdata_dir / 'author/0.0.1/test_3_md_hand_edited'
    test_template = testdata_dir / 'author/0.0.1/test_3_md_hand_edited/template.md'
    task_instance_folder = tmp_trestle_dir / 'test_task'

    (tmp_trestle_dir / '.trestle/author/test_task').mkdir(parents=True)
    task_template.touch()
    shutil.copy(test_template, task_template)
    shutil.copytree(test_template_folder, task_instance_folder)

    command_validate = 'trestle author docs validate -tn test_task'
    execute_command_and_assert(command_validate, 0, monkeypatch)

    command_validate = 'trestle author docs validate -tn test_task -gh Governed_Section'
    execute_command_and_assert(command_validate, 0, monkeypatch)

    # Change second section named Governed Section, should fail
    task_markdown = task_instance_folder / 'decisions_000.md'
    md_api = MarkdownAPI()
    header, tree = md_api.processor.process_markdown(task_markdown)

    # Make sure two nodes of the same name exists
    gov_nodes = tree.get_all_nodes_for_keys(['## Governed_Section'], True, False)

    assert len(gov_nodes) == 2

    # Change non-governed text
    tree.content.raw_text = tree.content.raw_text.replace('Not governed text', 'This is fine')
    md_api.write_markdown_with_header(task_markdown, header, tree.content.raw_text)

    command_string_validate_content = 'trestle author docs validate -tn test_task -gh Governed_Section'
    execute_command_and_assert(command_string_validate_content, 0, monkeypatch)

    # Now change governed stuff
    tree.content.raw_text = tree.content.raw_text.replace('Financial Approver', 'This is not fine')
    md_api.write_markdown_with_header(task_markdown, header, tree.content.raw_text)

    command_string_validate_content = 'trestle author docs validate -tn test_task -gh=Governed_Section'
    execute_command_and_assert(command_string_validate_content, 1, monkeypatch)

    # Now rename back
    tree.content.raw_text = tree.content.raw_text.replace('This is not fine', 'Financial Approver')
    md_api.write_markdown_with_header(task_markdown, header, tree.content.raw_text)

    command_string_validate_content = 'trestle author docs validate -tn test_task -gh=Governed_Section'
    execute_command_and_assert(command_string_validate_content, 0, monkeypatch)

    # Rename other governed section
    tree.content.raw_text = tree.content.raw_text.replace('Another_Governed_Section', 'Renamed Governed Section')
    md_api.write_markdown_with_header(task_markdown, header, tree.content.raw_text)

    command_string_validate_content = 'trestle author docs validate -tn test_task -gh=Another_Governed_Section'
    execute_command_and_assert(command_string_validate_content, 1, monkeypatch)

    # Validate non existing governed section
    command_string_validate_content = 'trestle author docs validate -tn test_task -gh=NonExisting'
    execute_command_and_assert(command_string_validate_content, 1, monkeypatch)



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Function to load and validate files while avoiding circular imports."""
import argparse
import logging
from pathlib import Path
from typing import Optional, Tuple

from trestle.common import const
from trestle.common.common_types import TG, TopLevelOscalModel
from trestle.common.file_utils import FileContentType
from trestle.common.model_utils import ModelUtils
from trestle.core.validator import Validator
from trestle.core.validator_factory import validator_factory

logger = logging.getLogger(__name__)


def load_validate_model_path(trestle_root: Path, model_path: Path) -> TopLevelOscalModel:
    """Load a model by its path and validate it."""
    _, _, model = ModelUtils.load_distributed(model_path, trestle_root)
    args = argparse.Namespace(mode=const.VAL_MODE_ALL, quiet=True)
    validator: Validator = validator_factory.get(args)
    if not validator.model_is_valid(model, True):
        logger.warning(f'Model loaded at {model_path} fails validation, but is being loaded anyway.')
    return model


def load_validate_model_name(
    trestle_root: Path,
    model_name: str,
    model_class: TG,
    file_content_type: Optional[FileContentType] = None
) -> Tuple[TG, Path]:
    """Load a model by its name and type and validate it."""
    model_path = ModelUtils.path_for_top_level_model(trestle_root, model_name, model_class, file_content_type)
    model = load_validate_model_path(trestle_root, model_path)
    return model, model_path



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2021 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Trestle Base Model."""
from typing import Any, Type, TypeVar

from pydantic import BaseModel, ValidationError

from trestle.common.err import TrestleError

Model = TypeVar('Model', bound='BaseModel')


class TrestleBaseModel(BaseModel):
    """Trestle Base Model. Serves as wrapper around BaseModel for overriding methods."""

    @classmethod
    def parse_obj(cls: Type['Model'], obj: Any) -> 'Model':
        """Parse object to the given class."""
        try:
            return super().parse_obj(obj)
        except ValidationError as e:
            # check if failed due to the wrong OSCAL version:
            oscal_version_error = False
            for err in e.errors():
                for field in err['loc']:
                    if field == 'oscal-version':
                        message = err['msg']
                        oscal_version_error = True
                        break
            if oscal_version_error:
                raise TrestleError(f'{message}')
            else:
                raise



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for trestle file content type module."""
from pathlib import Path

import pytest

from trestle.common.err import TrestleError
from trestle.core.models.file_content_type import FileContentType


def test_to_content_type() -> None:
    """Test to_content_type method."""
    assert FileContentType.to_content_type('.json') == FileContentType.JSON
    assert FileContentType.to_content_type('.yaml') == FileContentType.YAML
    assert FileContentType.to_content_type('') == FileContentType.DIRLIKE

    with pytest.raises(TrestleError):
        FileContentType.to_content_type('.invalid')


def test_to_file_extension() -> None:
    """Test to_file_extension method."""
    assert FileContentType.to_file_extension(FileContentType.JSON) == '.json'
    assert FileContentType.to_file_extension(FileContentType.YAML) == '.yaml'

    with pytest.raises(TrestleError):
        FileContentType.to_file_extension(-1)


def test_path_to_file_content_type(tmp_trestle_dir: Path) -> None:
    """Test path_to_file_content_type method."""
    tmp_stem = tmp_trestle_dir / 'content_type_test'

    tmp_file = tmp_stem.with_suffix('.json')
    tmp_file.touch()
    assert FileContentType.JSON == FileContentType.path_to_content_type(tmp_file)
    assert FileContentType.path_to_file_extension(tmp_file) == '.json'
    tmp_file.unlink()

    tmp_file = tmp_stem.with_suffix('.yaml')
    tmp_file.touch()
    assert FileContentType.YAML == FileContentType.path_to_content_type(tmp_file)
    assert FileContentType.path_to_file_extension(tmp_file) == '.yaml'
    tmp_file.unlink()

    tmp_file = tmp_stem.with_suffix('.yml')
    tmp_file.touch()
    assert FileContentType.YAML == FileContentType.path_to_content_type(tmp_file)
    assert FileContentType.path_to_file_extension(tmp_file) == '.yml'
    tmp_file.unlink()

    assert FileContentType.UNKNOWN == FileContentType.path_to_content_type(tmp_file)
    assert FileContentType.path_to_file_extension(tmp_file) == ''


def test_dir_to_content_type(tmp_path: Path) -> None:
    """Test failure of dir to file content type."""
    assert FileContentType.UNKNOWN == FileContentType.dir_to_content_type(tmp_path)



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Core constants module containing all constants."""

import string

FIELDS_SET = '__fields_set__'

ROOT = '__root__'

PACKAGE_OSCAL = 'trestle.oscal'

TRESTLE_CONFIG_DIR = '.trestle'
TRESTLE_DIST_DIR = 'dist'
TRESTLE_CONFIG_FILE = 'config.ini'
TRESTLE_KEEP_FILE = '.keep'
DRAWIO_FILE_EXT = '.drawio'
MARKDOWN_FILE_EXT = '.md'

ALLOWED_EXTENSIONS_IN_DIRS = {'.json', '.xml', '.yaml', '.yml', '.md'}

# these are hyphenated - no underscore - and mixed singular and plural
MODEL_TYPE_A_PLAN = 'assessment-plan'
MODEL_TYPE_A_RESULT = 'assessment-results'
MODEL_TYPE_CATALOG = 'catalog'
MODEL_TYPE_COMPDEF = 'component-definition'
MODEL_TYPE_POAM = 'plan-of-action-and-milestones'
MODEL_TYPE_PROFILE = 'profile'
MODEL_TYPE_SSP = 'system-security-plan'

# these are all plural and hyphenated - no underscore
MODEL_DIR_A_PLAN = 'assessment-plans'
MODEL_DIR_A_RESULT = 'assessment-results'
MODEL_DIR_CATALOG = 'catalogs'
MODEL_DIR_COMPDEF = 'component-definitions'
MODEL_DIR_POAM = 'plan-of-action-and-milestones'
MODEL_DIR_PROFILE = 'profiles'
MODEL_DIR_SSP = 'system-security-plans'

# these are singular except for assessment_results - and component, poam, and ssp are all abbreviated
# underscores here instead of hyphens
MODULE_NAME_A_PLAN = 'assessment_plan'
MODULE_NAME_A_RESULT = 'assessment_results'
MODULE_NAME_CATALOG = 'catalog'
MODULE_NAME_COMPDEF = 'component'
MODULE_NAME_POAM = 'poam'
MODULE_NAME_PROFILE = 'profile'
MODULE_NAME_SSP = 'ssp'

MODEL_MODULE_A_PLAN = f'{PACKAGE_OSCAL}.{MODULE_NAME_A_PLAN}'
MODEL_MODULE_A_RESULT = f'{PACKAGE_OSCAL}.{MODULE_NAME_A_RESULT}'
MODEL_MODULE_CATALOG = f'{PACKAGE_OSCAL}.{MODULE_NAME_CATALOG}'
MODEL_MODULE_COMPDEF = f'{PACKAGE_OSCAL}.{MODULE_NAME_COMPDEF}'
MODEL_MODULE_POAM = f'{PACKAGE_OSCAL}.{MODULE_NAME_POAM}'
MODEL_MODULE_PROFILE = f'{PACKAGE_OSCAL}.{MODULE_NAME_PROFILE}'
MODEL_MODULE_SSP = f'{PACKAGE_OSCAL}.{MODULE_NAME_SSP}'

MODEL_TYPE_LIST = [
    MODEL_TYPE_A_PLAN,
    MODEL_TYPE_A_RESULT,
    MODEL_TYPE_CATALOG,
    MODEL_TYPE_COMPDEF,
    MODEL_TYPE_POAM,
    MODEL_TYPE_PROFILE,
    MODEL_TYPE_SSP
]

MODEL_DIR_LIST = [
    MODEL_DIR_A_PLAN,
    MODEL_DIR_A_RESULT,
    MODEL_DIR_CATALOG,
    MODEL_DIR_COMPDEF,
    MODEL_DIR_POAM,
    MODEL_DIR_PROFILE,
    MODEL_DIR_SSP
]

MODEL_MODULE_LIST = [
    MODEL_MODULE_A_PLAN,
    MODEL_MODULE_A_RESULT,
    MODEL_MODULE_CATALOG,
    MODEL_MODULE_COMPDEF,
    MODEL_MODULE_POAM,
    MODEL_MODULE_PROFILE,
    MODEL_MODULE_SSP
]
"""Map of plural form of a model type to the oscal module that contains the classes related to it."""
MODEL_DIR_TO_MODEL_MODULE = {
    MODEL_DIR_A_PLAN: MODEL_MODULE_A_PLAN,
    MODEL_DIR_A_RESULT: MODEL_MODULE_A_RESULT,
    MODEL_DIR_CATALOG: MODEL_MODULE_CATALOG,
    MODEL_DIR_COMPDEF: MODEL_MODULE_COMPDEF,
    MODEL_DIR_POAM: MODEL_MODULE_POAM,
    MODEL_DIR_PROFILE: MODEL_MODULE_PROFILE,
    MODEL_DIR_SSP: MODEL_MODULE_SSP
}
"""Map of model type to oscal module."""
MODEL_TYPE_TO_MODEL_MODULE = {
    MODEL_TYPE_A_PLAN: MODEL_MODULE_A_PLAN,
    MODEL_TYPE_A_RESULT: MODEL_MODULE_A_RESULT,
    MODEL_TYPE_CATALOG: MODEL_MODULE_CATALOG,
    MODEL_TYPE_COMPDEF: MODEL_MODULE_COMPDEF,
    MODEL_TYPE_POAM: MODEL_MODULE_POAM,
    MODEL_TYPE_PROFILE: MODEL_MODULE_PROFILE,
    MODEL_TYPE_SSP: MODEL_MODULE_SSP
}
"""Map of model module to model type."""
MODEL_MODULE_TO_MODEL_TYPE = {
    MODEL_MODULE_A_PLAN: MODEL_TYPE_A_PLAN,
    MODEL_MODULE_A_RESULT: MODEL_TYPE_A_RESULT,
    MODEL_MODULE_CATALOG: MODEL_TYPE_CATALOG,
    MODEL_MODULE_COMPDEF: MODEL_TYPE_COMPDEF,
    MODEL_MODULE_POAM: MODEL_TYPE_POAM,
    MODEL_MODULE_PROFILE: MODEL_TYPE_PROFILE,
    MODEL_MODULE_SSP: MODEL_TYPE_SSP
}
"""Map of model type to model directory."""
MODEL_TYPE_TO_MODEL_DIR = {
    MODEL_TYPE_A_PLAN: MODEL_DIR_A_PLAN,
    MODEL_TYPE_A_RESULT: MODEL_DIR_A_RESULT,
    MODEL_TYPE_CATALOG: MODEL_DIR_CATALOG,
    MODEL_TYPE_COMPDEF: MODEL_DIR_COMPDEF,
    MODEL_TYPE_POAM: MODEL_DIR_POAM,
    MODEL_TYPE_PROFILE: MODEL_DIR_PROFILE,
    MODEL_TYPE_SSP: MODEL_DIR_SSP
}
"""Element path separator"""
ALIAS_PATH_SEPARATOR: str = '.'

# argument names
ARG_FILE = 'file'
ARG_FILE_SHORT = 'f'

ARG_ELEMENT = 'element'
ARG_ELEMENT_SHORT = 'e'

# argument descriptions
ARG_DESC_FILE = 'Path of the file'
ARG_DESC_ELEMENT = 'Path of the element in the OSCAL model'

# Index separater for naming directories representing collection properties
IDX_SEP = '__'

# Digit prefix for split files
FILE_DIGIT_PREFIX_LENGTH = 5

# Wildcard that can be used in the element path to represent all elements
ELEMENT_WILDCARD = '*'
ARG_VALIDATE = 'validate'
ARG_VALIDATE_SHORT = 'v'

ARG_MODE = 'mode'
ARG_MODE_SHORT = 'm'
ARG_DESC_MODE = 'Mode of the operation'

ARG_ITEM = 'item'
ARG_ITEM_SHORT = 'i'
ARG_DESC_ITEM = 'Item used'

VAL_MODE_CATALOG = 'catalog'
VAL_MODE_DUPLICATES = 'duplicates'
VAL_MODE_LINKS = 'links'
VAL_MODE_REFS = 'refs'
VAL_MODE_ALL = 'all'

IOF_SHORT = '-iof'
IOF_LONG = '--include-optional-fields'
IOF_HELP = 'Include fields that are optional in the OSCAL model when generating the new object.'

INIT_FULL_SHORT = '-fl'
INIT_FULL_LONG = '--full'
INIT_FULL_HELP = 'Initializes Trestle workspace for local, API and governed documents usage.'

INIT_GOVDOCS_SHORT = '-gd'
INIT_GOVDOCS_LONG = '--govdocs'
INIT_GOVDOCS_HELP = 'Initializes Trestle workspace for governed documents usage only.'

INIT_LOCAL_SHORT = '-loc'
INIT_LOCAL_LONG = '--local'
INIT_LOCAL_HELP = 'Initializes Trestle workspace for local management of OSCAL models.'

FILE_ENCODING = 'utf8'

# Trestle documentation

WEBSITE_ROOT = 'https://ibm.github.io/compliance-trestle'

BUG_REPORT = 'https://github.com/IBM/compliance-trestle/issues/new/choose'

# Sample objects
SAMPLE_UUID_STR = 'A0000000-0000-4000-8000-000000000000'

WINDOWS_PLATFORM_STR = 'Windows'

# constants related to cache

TRESTLE_CACHE_DIR = TRESTLE_CONFIG_DIR + '/cache'

HOUR_SECONDS: int = 3600

DAY_SECONDS: int = 24 * HOUR_SECONDS

FILE_URI = 'file:///'

SFTP_URI = 'sftp://'

HTTPS_URI = 'https://'

# extracts letter, colon and single slash
WINDOWS_DRIVE_URI_REGEX = r'([A-Za-z]:[\\/]?)[^\\/]'

WINDOWS_DRIVE_LETTER_REGEX = r'[A-Za-z]:'

# extract two groups corresponding to prop name and value
# need to strip white space from end of value
PROPERTY_REGEX = r'(?:###\s+Property\s+)([^\s]*)\s*:\s+(.*)'

PART_REGEX = r'(?:##\s+Part\s+)(.*)'

CONTROL_REGEX = r'(?:##\s+Control\s+)(.*)'

AFTER_HASHES_REGEX = r'(?:##*\s+)(.*)'

CACHE_ABS_DIR = '__abs__'

UNIX_CACHE_ROOT = '__root__'

TRESTLE_HREF_HEADING = 'trestle://'

TRESTLE_HREF_REGEX = '^trestle://[^/]'

MATCH_ALL_EXCEPT_LETTERS_UNDERSCORE_SPACE_REGEX = '[^a-zA-Z0-9-_ \n]'

# extracts foo and bar from ...[foo](bar)...
MARKDOWN_URL_REGEX = r'\[([^\]]+)\]\(([^)]+)\)'

# Governed header template version
TEMPLATE_VERSION_REGEX = r'[0-9]+.[0-9]+.[0-9]+'

OBJECTIVE_PART = 'objective'
TABLE_OF_PARAMS_PART = 'table_of_parameters'

# extracts standalone uuid's from anywhere in string
UUID_REGEX = r'(?:^|[0-9A-Za-f])([0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-4[0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12})(?:$|[^0-9A-Za-z])'  # noqa FS003 E501

SSP_MD_HRULE_LINE = '______________________________________________________________________'

SSP_MD_IMPLEMENTATION_QUESTION = 'What is the solution and how is it implemented?'

SSP_MD_LEAVE_BLANK_TEXT = '<!-- Please leave this section blank and enter implementation details in the parts below. -->'  # noqa E501

SSP_ADD_IMPLEMENTATION_PREFIX = 'Add control implementation description here for '

STATEMENT = 'statement'

SSP_ADD_IMPLEMENTATION_FOR_STATEMENT_TEXT = SSP_ADD_IMPLEMENTATION_PREFIX + STATEMENT

SSP_ADD_IMPLEMENTATION_FOR_ITEM_TEXT = SSP_ADD_IMPLEMENTATION_PREFIX + 'item'

SSP_ADD_IMPLEMENTATION_FOR_CONTROL_TEXT = SSP_ADD_IMPLEMENTATION_PREFIX + 'control'

SSP_SYSTEM_CONTROL_IMPLEMENTATION_TEXT = 'This is the control implementation for the system.'

PROFILE_ADD_REQUIRED_SECTION_FOR_CONTROL_TEXT = 'Add prose here for required Section'

NCNAME_REGEX = r'^[_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD][_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD\-\.0-9\u00B7\u0300-\u036F\u203F-\u2040]*$'  # noqa FS003 E501

NCNAME_UTF8_FIRST_CHAR_OPTIONS = string.ascii_letters + '_'

NCNAME_UTF8_OTHER_CHAR_OPTIONS = string.ascii_letters + string.digits + '_.-'

DISPLAY_VERBOSE_OUTPUT = 'Display verbose output'

HELP_YAML_PATH = 'Path to the optional yaml header file'

HELP_OVERWRITE_HEADER_VALUES = (
    'Flag to overwrite values in a markdown control header.'
    + ' If a separate yaml header is passed in with -y, any items in the markdown header that are common with the'
    + ' provided header will be overwritten by the new values.  But new items passed in will always be added to the'
    + ' markdown header.'
)

HELP_REGENERATE = 'Flag to force generation of new uuids in the model'

HELP_VERSION = 'New version for the assembled model'

HELP_SECTIONS = 'Comma-separated list of sections as short_name_no_spaces:long name with spaces'

HELP_REQUIRED_SECTIONS = 'Short names of sections that must be in the assembled model, comma-separated'

HELP_ALLOWED_SECTIONS = 'Short names of sections that are allowed to be in the assembled model, comma-separated'

HELP_MARKDOWN_NAME = 'Name of the output generated profile markdown folder'

FILTER_BY_PROFILE = 'filter-by-profile'

FILTER_BY_COMPONENTS = 'filter-by-components'

FILTER_EXCLUDE_COMPONENTS = 'filter-exclude-components'

GENERATE_RESOLVED_CATALOG = 'generate-resolved-catalog'

TRANSFORM_TYPES = [FILTER_BY_PROFILE, FILTER_BY_COMPONENTS, FILTER_EXCLUDE_COMPONENTS, GENERATE_RESOLVED_CATALOG]

SSP_MAIN_COMP_NAME = 'This System'

TRESTLE_TAG = 'x-trestle-'

TRESTLE_PROPS_TAG = TRESTLE_TAG + 'props'

NAMESPACE_NIST = 'https://csrc.nist.gov/ns/oscal'

SSP_FEDRAMP_TAG = TRESTLE_TAG + 'fedramp-props'

TRESTLE_GLOBAL_TAG = TRESTLE_TAG + 'global'

PROFILE_TITLE = 'profile-title'

NAMESPACE_FEDRAMP = 'https://fedramp.gov/ns/oscal'

LEV_AUTH_UUID = 'leveraged-authorization-uuid'

STATUS_INHERITED = 'inherited'

STATUS_PARTIALLY_IMPLEMENTED = 'partially-implemented'

STATUS_PLANNED_COMPLETION_DATE = 'planned-completion-date'

STATUS_COMPLETION_DATE = 'completion-date'

CONTROL_ORIGINATION = 'control-origination'

IMPLEMENTATION_STATUS = 'implementation-status'

IMPLEMENTATION_STATUS_HEADER = 'Implementation Status'

IMPLEMENTATION_STATUS_REMARKS_HEADER = 'Implementation Status Remarks'

REMARKS = 'Remarks'

STATUS_REMARKS = 'status-remarks'

# Following 5 are allowed state tokens for
# SSP -> ControlImplementation -> ImplementedRequirements -> ByComponents -> common.ImplementationStatus -> State
# Also                         -> ImplementedRequirements -> Statements -> ByComponents ...
# But NIST says they may also be locally defined
STATUS_IMPLEMENTED = 'implemented'

STATUS_PARTIAL = 'partial'

STATUS_PARTIALLY_IMPLEMENTED = 'partially-implemented'

STATUS_PLANNED = 'planned'

STATUS_ALTERNATIVE = 'alternative'

STATUS_NOT_APPLICABLE = 'not-applicable'

# Following 4 needed by SSP -> SystemImplementation -> SystemComponent -> Status -> State1
# and by SSP -> SystemCharacteristics -> Status1 -> State
STATUS_OPERATIONAL = 'operational'

STATUS_UNDER_DEVELOPMENT = 'under-development'

STATUS_DISPOSITION = 'disposition'

STATUS_OTHER = 'other'

# Needed only by SystemCharacteristics
STATUS_UNDER_MAJOR_MODIFICATION = 'under-major-modification'

STATUS_ALL = [
    STATUS_IMPLEMENTED,
    STATUS_PARTIAL,
    STATUS_PARTIALLY_IMPLEMENTED,
    STATUS_PLANNED,
    STATUS_ALTERNATIVE,
    STATUS_NOT_APPLICABLE,
    STATUS_OPERATIONAL,
    STATUS_UNDER_DEVELOPMENT,
    STATUS_DISPOSITION,
    STATUS_OTHER,
    STATUS_UNDER_MAJOR_MODIFICATION
]

STATUS_PROMPT = f'<!-- For implementation status enter one of: {STATUS_IMPLEMENTED}, {STATUS_PARTIAL}, {STATUS_PLANNED}, {STATUS_ALTERNATIVE}, {STATUS_NOT_APPLICABLE} -->'  # noqa E501

RULES_WARNING = '<!-- Note that the list of rules under ### Rules: is read-only and changes will not be captured after assembly to JSON -->'  # noqa E501

RESPONSIBLE_ROLE = 'responsible-role'

RESPONSIBLE_ROLES = 'responsible-roles'

HELP_SET_PARAMS = 'set parameters and properties based on the yaml header in control markdown'

SET_PARAMS_TAG = TRESTLE_TAG + 'set-params'

RULES_PARAMS_TAG = TRESTLE_TAG + 'rules-params'

COMP_DEF_RULES_PARAM_VALS_TAG = TRESTLE_TAG + 'comp-def-rules-param-vals'

PARAM_VALUES_TAG = TRESTLE_TAG + 'param-values'

COMP_DEF_RULES_TAG = TRESTLE_TAG + 'comp-def-rules'

PROFILE_VALUES = 'profile-values'

COMP_DEF_VALUES = 'comp-def-values'

VALUES = 'values'

SECTIONS_TAG = TRESTLE_TAG + 'sections'

EDITABLE_CONTENT = 'Editable Content'

SORT_ID = 'sort-id'

TRESTLE_IMP_STATUS_TAG = 'trestle-imp-status'

TRESTLE_ADD_PROPS_TAG = TRESTLE_TAG + 'add-props'

TRESTLE_INHERITED_PROPS_TAG = TRESTLE_TAG + 'inherited-props'

CONTROL_OBJECTIVE_HEADER = '## Control Objective'

CONTROL_HEADER = '## Control'

REPLACE_ME = 'REPLACE_ME'

YAML_PROPS_COMMENT = """  # Add or modify control properties here
  # Properties may be at the control or part level
  # Add control level properties like this:
  #   - name: ac1_new_prop
  #     value: new property value
  #
  # Add properties to a statement part like this, where "b." is the label of the target statement part
  #   - name: ac1_new_prop
  #     value: new property value
  #     smt-part: b.
  #
"""

DISPLAY_NAME = 'display-name'

TRESTLE_GENERIC_NS = 'https://ibm.github.io/compliance-trestle/schemas/oscal'

RESOLUTION_SOURCE = 'resolution-source'

# call it tracker to distinguish from the externally visible TRESTLE_INHERITED_PROPS_TAG
TRESTLE_INHERITED_PROPS_TRACKER = 'trestle_inherited_props_tracker'

RULE_ID = 'Rule_Id'

RULE_DESCRIPTION = 'Rule_Description'

PARAMETER_ID = 'Parameter_Id'

PARAMETER_DESCRIPTION = 'Parameter_Description'

PARAMETER_VALUE_ALTERNATIVES = 'Parameter_Value_Alternatives'



# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Generic classes to support both SSP and DefinedComponents."""
from __future__ import annotations

import copy
from typing import List, Optional
from uuid import uuid4

from pydantic import Field, constr

import trestle.oscal.component as comp
import trestle.oscal.ssp as ossp
from trestle.common import const
from trestle.common.list_utils import as_list
from trestle.core.control_interface import ControlInterface
from trestle.core.trestle_base_model import TrestleBaseModel
from trestle.oscal import common


class GenericByComponent(TrestleBaseModel):
    """Generic ByComponent for SSP and DefinedComponent."""

    # only in SSP
    component_uuid: constr(
        regex=r'^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-4[0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$'  # noqa FS003
    ) = Field(
        ...,
        alias='component_uuid',
        description='A machine-oriented identifier reference to the component that is implemeting a given control.',
        title='Component Universally Unique Identifier Reference',
    )
    uuid: constr(
        regex=r'^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-4[0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$'  # noqa FS003
    ) = Field(
        ...,
        description=  # noqa E251
        'A machine-oriented, globally unique identifier with cross-instance scope that can be used to reference this by-component entry elsewhere in this or other OSCAL instances. The locally defined UUID of the by-component entry can be used to reference the data item locally or globally (e.g., in an imported OSCAL instance). This UUID should be assigned per-subject, which means it should be consistently used to identify the same subject across revisions of the document.',  # noqa E501
        title='By-Component Universally Unique Identifier',
    )
    description: str = Field(
        ...,
        description=  # noqa E251
        'An implementation statement that describes how a control or a control statement is implemented within the referenced system component.',  # noqa E501
        title='Control Implementation Description',
    )
    props: Optional[List[common.Property]] = Field(None)
    links: Optional[List[common.Link]] = Field(None)
    set_parameters: Optional[List[GenericSetParameter]] = Field(None, alias='set-parameters')
    implementation_status: Optional[common.ImplementationStatus] = Field(None, alias='implementation-status')
    # removed export, inherited, satisfied
    responsible_roles: Optional[List[common.ResponsibleRole]] = Field(None, alias='responsible-roles')
    remarks: Optional[common.Remarks] = None

    @staticmethod
    def generate() -> GenericByComponent:
        """Generate instance of generic ByComponent."""
        uuid = str(uuid4())
        return GenericByComponent(component_uuid=const.SAMPLE_UUID_STR, uuid=uuid, description=const.REPLACE_ME)


class GenericStatement(TrestleBaseModel):
    """Generic statement for SSP and DefinedComp."""

    statement_id: constr(
        regex=  # noqa E251
        r'^[_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD][_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD\-\.0-9\u00B7\u0300-\u036F\u203F-\u2040]*$'  # noqa FS003 E501
    ) = Field(
        ...,
        alias='statement_id',
        description='A human-oriented identifier reference to a control statement.',
        title='Control Statement Reference',
    )
    uuid: constr(
        regex=r'^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-4[0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$'  # noqa FS003 F722
    ) = Field(
        ...,
        description=  # noqa E251
        'A machine-oriented, globally unique identifier with cross-instance scope that can be used to reference this control statement elsewhere in this or other OSCAL instances. The UUID of the control statement in the source OSCAL instance is sufficient to reference the data item locally or globally (e.g., in an imported OSCAL instance).',  # noqa E501
        title='Control Statement Reference Universally Unique Identifier',
    )
    # this is not in ssp statement
    description: str = Field(
        ...,
        description='A summary of how the containing control statement is implemented by the component or capability.',
        title='Statement Implementation Description',
    )
    props: Optional[List[common.Property]] = Field(None)
    links: Optional[List[common.Link]] = Field(None)
    responsible_roles: Optional[List[common.ResponsibleRole]] = Field(None, alias='responsible-roles')
    remarks: Optional[common.Remarks] = None
    # ssp has following
    by_components: Optional[List[GenericByComponent]] = Field(None, alias='by-components')

    @staticmethod
    def generate() -> GenericStatement:
        """Generate instance of GenericStatement."""
        uuid = str(uuid4())
        return GenericStatement(statement_id=const.REPLACE_ME, uuid=uuid, description=const.REPLACE_ME)

    def as_comp_def(self) -> List[comp.Statement]:
        """Represent in comp_def form."""
        # convert all by_comps to statements with status
        statements: List[comp.Statement] = []
        for by_comp in as_list(self.by_components):
            stat_dict = copy.deepcopy(self.__dict__)
            stat_dict.pop('by_components', None)
            stat_dict['description'] = by_comp.description
            new_stat = comp.Statement(**stat_dict)
            ControlInterface.insert_status_in_props(new_stat, by_comp.implementation_status)
            statements.append(new_stat)
        return statements

    def as_ssp(self) -> ossp.Statement:
        """Represent in ssp form."""
        class_dict = copy.deepcopy(self.__dict__)
        class_dict.pop('description', None)
        return ossp.Statement(**class_dict)


class GenericComponent(TrestleBaseModel):
    """Generic component for SSP SystemComponent and DefinedComponent."""

    uuid: constr(
        regex=r'^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-4[0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$'  # noqa FS003 F722
    ) = Field(
        ...,
        description=  # noqa E251
        'A machine-oriented, globally unique identifier with cross-instance scope that can be used to reference this component elsewhere in this or other OSCAL instances. The locally defined UUID of the component can be used to reference the data item locally or globally (e.g., in an imported OSCAL instance). This UUID should be assigned per-subject, which means it should be consistently used to identify the same subject across revisions of the document.',  # noqa E501
        title='Component Identifier',
    )
    type: constr(regex=r'^\S(.*\S)?$') = Field(  # noqa A003 F722
        ...,
        description='A category describing the purpose of the component.',
        title='Component Type',
    )
    title: str = Field(
        ...,
        description='A human readable name for the component.',
        title='Component Title',
    )
    description: str = Field(
        ...,
        description='A description of the component, including information about its function.',
        title='Component Description',
    )
    purpose: Optional[str] = Field(
        None,
        description='A summary of the technological or business purpose of the component.',
        title='Purpose',
    )
    props: Optional[List[common.Property]] = Field(None)
    links: Optional[List[common.Link]] = Field(None)
    responsible_roles: Optional[List[common.ResponsibleRole]] = Field(None, alias='responsible-roles')
    protocols: Optional[List[common.Protocol]] = Field(None)
    # ssp does not have
    control_implementations: Optional[List[GenericControlImplementation]] = Field(None, alias='control-implementations')
    remarks: Optional[common.Remarks] = None
    # ssp has
    status: common.ImplementationStatus

    def as_defined_component(self) -> comp.DefinedComponent:
        """Convert to DefinedComponent."""
        status = self.status
        class_dict = copy.deepcopy(self.__dict__)
        class_dict.pop('status', None)
        def_comp = comp.DefinedComponent(**class_dict)
        ControlInterface.insert_status_in_props(def_comp, status)
        return def_comp

    @classmethod
    def from_defined_component(cls, def_comp: comp.DefinedComponent) -> GenericComponent:
        """Convert defined component to generic."""
        status = ControlInterface.get_status_from_props(def_comp)
        class_dict = copy.deepcopy(def_comp.__dict__)
        class_dict['status'] = status
        return cls(**class_dict)

    def as_system_component(self) -> ossp.SystemComponent:
        """Convert to SystemComponent."""
        class_dict = copy.deepcopy(self.__dict__)
        class_dict.pop('control_implementations', None)
        status_str = self.status.state if self.status else 'other'
        if status_str not in ['under-development', 'operational', 'disposition', 'other']:
            status_str = 'other'
        class_dict['status'] = ossp.Status(state=ossp.State1(status_str), remarks=self.status.remarks)
        return ossp.SystemComponent(**class_dict)

    @classmethod
    def from_system_component(cls, sys_comp: ossp.SystemComponent) -> GenericComponent:
        """Convert system component to generic."""
        class_dict = copy.deepcopy(sys_comp.__dict__)
        class_dict.pop('control_implementations', None)
        status_str = 'other'
        if 'status' in class_dict:
            status_str = class_dict['status'].state.name
        class_dict['status'] = common.ImplementationStatus(state=status_str)
        return cls(**class_dict)

    @staticmethod
    def generate() -> GenericComponent:
        """Generate instance of GenericComponent."""
        uuid = str(uuid4())
        status = common.ImplementationStatus(state=const.STATUS_PLANNED)
        return GenericComponent(
            uuid=uuid, type=const.REPLACE_ME, title=const.REPLACE_ME, description=const.REPLACE_ME, status=status
        )


class GenericSetParameter(TrestleBaseModel):
    """Generic SetParameter for SSP and DefinedComponent."""

    param_id: constr(
        regex=  # noqa E251
        r'^[_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD][_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD\-\.0-9\u00B7\u0300-\u036F\u203F-\u2040]*$'  # noqa E501
    ) = Field(
        ...,
        alias='param-id',
        description=  # noqa E251
        "A human-oriented reference to a parameter within a control, who's catalog has been imported into the current implementation context.",  # noqa E501
        title='Parameter ID',
    )
    values: List[common.Value] = Field(...)
    remarks: Optional[common.Remarks] = None


class GenericImplementedRequirement(TrestleBaseModel):
    """Generic ImplementedRequirement for SSP and DefinedComponent."""

    uuid: constr(
        regex=r'^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-4[0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$'  # noqa FS003 F722
    ) = Field(
        ...,
        description=  # noqa E251
        'A machine-oriented, globally unique identifier with cross-instance scope that can be used to reference a specific control implementation elsewhere in this or other OSCAL instances. The locally defined UUID of the control implementation can be used to reference the data item locally or globally (e.g., in an imported OSCAL instance).This UUID should be assigned per-subject, which means it should be consistently used to identify the same subject across revisions of the document.',  # noqa E501
        title='Control Implementation Identifier',
    )
    control_id: constr(
        regex=  # noqa E251
        r'^[_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD][_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD\-\.0-9\u00B7\u0300-\u036F\u203F-\u2040]*$'  # noqa E501
    ) = Field(
        ...,
        alias='control-id',
        description=  # noqa E251
        'A human-oriented identifier reference to a control with a corresponding id value. When referencing an externally defined control, the Control Identifier Reference must be used in the context of the external / imported OSCAL instance (e.g., uri-reference).',  # noqa E501
        title='Control Identifier Reference',
    )
    # only compdef has description
    description: str = Field(
        ...,
        description=  # noqa E251
        'A description of how the specified control is implemented for the containing component or capability.',  # noqa E501
        title='Control Implementation Description',
    )
    props: Optional[List[common.Property]] = Field(None)
    links: Optional[List[common.Link]] = Field(None)
    set_parameters: Optional[List[GenericSetParameter]] = Field(None, alias='set-parameters')
    responsible_roles: Optional[List[common.ResponsibleRole]] = Field(None, alias='responsible-roles')
    statements: Optional[List[GenericStatement]] = Field(None)
    remarks: Optional[common.Remarks] = None
    # ssp has following
    by_components: Optional[List[GenericByComponent]] = Field(None, alias='by-components')

    @staticmethod
    def generate() -> GenericImplementedRequirement:
        """Generate instance of this class."""
        uuid = str(uuid4())
        class_dict = {'uuid': uuid, 'control-id': const.REPLACE_ME, 'description': const.REPLACE_ME}
        return GenericImplementedRequirement(**class_dict)

    def as_comp_def(self) -> comp.ImplementedRequirement:
        """Convert to defined component form."""
        class_dict = copy.deepcopy(self.__dict__)
        del class_dict['by_components']
        new_stat_list = []
        for statement in as_list(self.statements):
            new_stat_list.extend(statement.as_comp_def())
        if new_stat_list:
            class_dict['statements'] = new_stat_list
        return comp.ImplementedRequirement(**class_dict)

    def as_ssp(self) -> ossp.ImplementedRequirement:
        """Convert to ssp form."""
        class_dict = copy.deepcopy(self.__dict__)
        del class_dict['description']
        new_stat_list = []
        for statement in as_list(self.statements):
            new_stat_list.append(statement.as_ssp())
        if new_stat_list:
            class_dict['statements'] = new_stat_list
        return ossp.ImplementedRequirement(**class_dict)


class GenericControlImplementation(TrestleBaseModel):
    """Generic control implementation for SSP and CompDef."""

    # not in ssp
    uuid: constr(
        regex=r'^[0-9A-Fa-f]{8}-[0-9A-Fa-f]{4}-4[0-9A-Fa-f]{3}-[89ABab][0-9A-Fa-f]{3}-[0-9A-Fa-f]{12}$'  # noqa FS003 F722
    ) = Field(
        ...,
        description=  # noqa E251
        'A machine-oriented, globally unique identifier with cross-instance scope that can be used to reference a set of implemented controls elsewhere in this or other OSCAL instances. The locally defined UUID of the control implementation set can be used to reference the data item locally or globally (e.g., in an imported OSCAL instance). This UUID should be assigned per-subject, which means it should be consistently used to identify the same subject across revisions of the document.',  # noqa E501
        title='Control Implementation Set Identifier',
    )
    # not in ssp
    source: str = Field(
        ...,
        description=  # noqa E251
        'A reference to an OSCAL catalog or profile providing the referenced control or subcontrol definition.',  # noqa E501
        title='Source Resource Reference',
    )
    description: str = Field(
        ...,
        description=  # noqa E251
        'A description of how the specified set of controls are implemented for the containing component or capability.',  # noqa E501
        title='Control Implementation Description',
    )
    # not in ssp
    props: Optional[List[common.Property]] = Field(None)
    # not in ssp
    links: Optional[List[common.Link]] = Field(None)
    set_parameters: Optional[List[GenericSetParameter]] = Field(None, alias='set-parameters')
    implemented_requirements: List[GenericImplementedRequirement] = Field(..., alias='implemented-requirements')

    @staticmethod
    def generate() -> GenericControlImplementation:
        """Generate instance of this class."""
        uuid = str(uuid4())
        imp_reqs = [GenericImplementedRequirement.generate()]
        class_dict = {
            'uuid': uuid,
            'control-id': const.REPLACE_ME,
            'source': const.REPLACE_ME,
            'description': const.REPLACE_ME,
            'implemented-requirements': imp_reqs
        }
        return GenericControlImplementation(**class_dict)

    def as_ssp(self) -> ossp.ControlImplementation:
        """Represent in ssp form."""
        imp_reqs = []
        for imp_req in self.implemented_requirements:
            imp_reqs.append(imp_req.as_ssp())
        class_dict = self.__dict__
        for prop in ['uuid', 'source', 'props', 'links', 'implemented_requirements']:
            class_dict.pop(prop, None)
        if imp_reqs:
            class_dict['implemented-requirements'] = imp_reqs
        return ossp.ControlImplementation(**class_dict)



# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interfaces for use within other trestle functions defined as pydantic data models."""
from typing import Dict, Optional

from pydantic import Extra

import trestle.oscal.assessment_plan as o_ap
import trestle.oscal.assessment_results as o_ar
import trestle.oscal.catalog as o_catalog
import trestle.oscal.component as o_component
import trestle.oscal.poam as o_poam
import trestle.oscal.profile as o_profile
import trestle.oscal.ssp as o_ssp
from trestle.core.trestle_base_model import TrestleBaseModel


class OSCALAssembly(TrestleBaseModel):
    """Data model to represent an assembled set of OSCAL objects.

    Here the assembly represents the constraints as expected by the current OSCAL
    schema. At this point in time a 'flat' model has been chosen rather than an tree.
    """

    poam: Optional[o_poam.PlanOfActionAndMilestones] = None
    sar: Optional[o_ar.AssessmentResults] = None
    sap: Optional[o_ap.AssessmentPlan] = None
    ssp: Optional[o_ssp.SystemSecurityPlan] = None
    profiles: Optional[Dict[str, o_profile.Profile]] = None
    catalogs: Optional[Dict[str, o_catalog.Catalog]] = None
    components: Optional[Dict[str, o_component.ComponentDefinition]] = None

    class Config:
        """Pydantic config overrides."""

        allow_population_by_field_name = True
        # Enforce strict schema
        extra = Extra.forbid
        # Validate on assignment of variables to ensure no escapes
        validate_assignment = True



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Validate by confirming all refs have corresponding id."""
from trestle.common.common_types import TopLevelOscalModel
from trestle.common.list_utils import as_list
from trestle.core.validator import Validator


class RefsValidator(Validator):
    """Validator to confirm all references in responsible parties are found in roles."""

    def model_is_valid(self, model: TopLevelOscalModel, quiet: bool) -> bool:
        """
        Test if the model is valid.

        args:
            model: A top level OSCAL model.
            quiet: Don't report msgs unless invalid.

        returns:
            True (valid) if the model's responsible parties match those found in roles.
        """
        roles = as_list(model.metadata.roles)
        role_ids = [role.id for role in roles]
        responsible_parties = as_list(model.metadata.responsible_parties)
        if not responsible_parties:
            return True
        party_roles = [party.role_id for party in responsible_parties]
        # return true if all party roles are in the roles list
        return all(item in role_ids for item in party_roles)



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2021 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""AuthorCommonCommands - reusable utilities to increase code base abstraction for author command."""
import argparse
import logging
import pathlib

import trestle.common.log as log
from trestle.common import file_utils
from trestle.common.const import ARG_VALIDATE, TRESTLE_CONFIG_DIR
from trestle.common.err import TrestleError
from trestle.core.commands.author.consts import START_TEMPLATE_VERSION
from trestle.core.commands.author.versioning.template_versioning import TemplateVersioning
from trestle.core.commands.command_docs import CommandPlusDocs
from trestle.core.commands.common.return_codes import CmdReturnCodes

logger = logging.getLogger(__name__)


class AuthorCommonCommand(CommandPlusDocs):
    """Extension for the subset of commands that operate using the common mode structure."""

    trestle_root: pathlib.Path

    task_name: str

    def _initialize(self, args: argparse.Namespace) -> int:
        log.set_log_level_from_args(args)
        # Externalize
        self.trestle_root = args.trestle_root
        self.task_name = args.task_name

        try:
            self.global_ = args.__getattribute__('global')
        except AttributeError:
            self.global_ = None

        if self.task_name:
            self.task_path = self.trestle_root / self.task_name
            if not file_utils.is_directory_name_allowed(self.task_name):
                logger.error(
                    f'Task name {self.task_name} is invalid as it interferes with OSCAL and trestle reserved names.'
                )
                return CmdReturnCodes.COMMAND_ERROR.value

        rc = self._setup_template_dir(args)

        return rc

    def rel_dir(self, path: pathlib.Path) -> str:
        """Stringify a directory relative to trestle root."""
        return str(path.relative_to(self.trestle_root))

    def _setup_template_dir(self, args: argparse.Namespace) -> int:
        """Set template directory and update to new format."""
        if not self.global_ and self.task_name is None:
            logger.error('At least a global flag or a task name should be provided.')
            return CmdReturnCodes.INCORRECT_ARGS.value
        if self.global_:
            old_template_dir = self.trestle_root / TRESTLE_CONFIG_DIR / 'author' / '__global__'
            self._set_template_version_to_latest(args, old_template_dir)
            self.template_dir = old_template_dir / args.template_version
        elif self.task_name and not self.global_:
            old_template_dir = self.trestle_root / TRESTLE_CONFIG_DIR / 'author' / self.task_name
            self._set_template_version_to_latest(args, old_template_dir)
            self.template_dir = old_template_dir / args.template_version

        if old_template_dir.exists():
            TemplateVersioning.update_template_folder_structure(old_template_dir)

        return CmdReturnCodes.SUCCESS.value

    def _set_template_version_to_latest(self, args: argparse.Namespace, template_dir: pathlib.Path):
        """Set template version argument to the latest version if none was given."""
        if not TemplateVersioning.is_valid_version(args.template_version):
            raise TrestleError(f'Version {args.template_version} is invalid, version format should be: 0.0.1')
        if args.template_version is None and args.mode == ARG_VALIDATE:
            # in validate mode no version will validate instances based on header version
            args.template_version = ''
        if args.template_version is None:
            args.template_version = START_TEMPLATE_VERSION
            if template_dir.exists():
                all_versions = TemplateVersioning.get_all_versions_for_task(template_dir)
                if all_versions:
                    args.template_version = max(all_versions)
        if args.template_version == '':
            logger.info('Instances will be validated against template version specified in their headers.')
        else:
            logger.info(f'Set template version to {args.template_version}.')



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""TransformFactory and related tests."""

import pytest

import trestle.common.err as err
from trestle.core.base_model import OscalBaseModel
from trestle.oscal.common import RoleId
from trestle.transforms.results import Results
from trestle.transforms.transformer_factory import ResultsTransformer, TransformerBase
from trestle.transforms.transformer_singleton import transformer_factory as tf


class DummyTransformer(TransformerBase):
    """Example transformer."""

    def transform(self, blob: str) -> OscalBaseModel:
        """Transform the blob."""
        # This example uses RoleId because it is a simple string and is otherwise arbitrary
        return RoleId(__root__='my_string')


class DummyResultsTransformer(ResultsTransformer):
    """Example results transformer."""

    def transform(self, blob: str) -> Results:
        """Transform the blob."""
        return Results(__root__=[])


def test_basic_transformer_operations() -> None:
    """Test basic transformer operations."""
    tf.register_transformer('dummy', DummyTransformer)
    transformer: TransformerBase = tf.get('dummy')
    assert transformer.transform('foo') == RoleId(__root__='my_string')


def test_results_transformer() -> None:
    """Test results transformer."""
    tf.register_transformer('dummy', DummyResultsTransformer)
    transformer: ResultsTransformer = tf.get('dummy')
    assert transformer.transform('foo') == Results(__root__=[])


def test_transformer_not_registered() -> None:
    """Test basic transformer operations."""
    with pytest.raises(err.TrestleError):
        transformer: TransformerBase = tf.get('foo')
        assert transformer is not None



# Copyright (c) 2021 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the ssp_io."""
import pathlib
from typing import Tuple

from _pytest.monkeypatch import MonkeyPatch

from tests.test_utils import delete_line_in_file, execute_command_and_assert, setup_for_ssp
from tests.trestle.core.commands.author.ssp_test import insert_prose

from trestle.common import file_utils
from trestle.common.const import CONTROL_ORIGINATION, IMPLEMENTATION_STATUS, STATUS_INHERITED, STATUS_PLANNED
from trestle.core import profile_resolver
from trestle.core.commands.author.ssp import SSPGenerate
from trestle.core.markdown.markdown_node import MarkdownNode
from trestle.core.remote import cache
from trestle.core.ssp_io import SSPMarkdownWriter
from trestle.oscal.common import Property
from trestle.oscal.ssp import SystemSecurityPlan

prof_name = 'main_profile'
ssp_name = 'my_ssp'


def setup_test(tmp_trestle_dir: pathlib.Path, testdata_dir: pathlib.Path,
               trestle_root: str) -> Tuple[pathlib.Path, SystemSecurityPlan]:
    """Prepare ssp test."""
    profile_path = tmp_trestle_dir / f'profiles/{prof_name}/profile.json'
    new_catalog_dir = tmp_trestle_dir / f'catalogs/{prof_name}_resolved_catalog'
    new_catalog_dir.mkdir(parents=True, exist_ok=True)

    ssp_json = testdata_dir / 'author/ssp/ssp_example.json'
    fetcher = cache.FetcherFactory.get_fetcher(trestle_root, str(ssp_json))
    ssp_obj: SystemSecurityPlan
    ssp_obj, parent_alias = fetcher.get_oscal(True)
    ssp_obj.control_implementation.implemented_requirements[0].props = [
        Property(name=IMPLEMENTATION_STATUS, value=STATUS_PLANNED),
        Property(name=CONTROL_ORIGINATION, value=STATUS_INHERITED),
    ]

    assert parent_alias == 'system-security-plan'

    return profile_path, ssp_obj


def test_ssp_writer(testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test ssp writer from cli."""
    gen_args, _, _ = setup_for_ssp(True, True, tmp_trestle_dir, prof_name, ssp_name)

    profile_path, ssp_obj = setup_test(tmp_trestle_dir, testdata_dir, gen_args.trestle_root)

    resolved_catalog = profile_resolver.ProfileResolver.get_resolved_profile_catalog(tmp_trestle_dir, profile_path)
    ssp_writer = SSPMarkdownWriter(gen_args.trestle_root)
    ssp_writer.set_catalog(resolved_catalog)
    md_text = ssp_writer.get_control_statement('ac-2', 1)
    assert md_text is not None

    ssp_writer.set_ssp(ssp_obj)
    roles_md = ssp_writer.get_responsible_roles_table('ac-2', 1)
    assert roles_md

    md_text1 = ssp_writer._parameter_table('ac-2', 1)
    assert md_text1

    md_text2 = ssp_writer._parameter_table('ac-2', 1, True)
    assert md_text2
    assert '| Label or Choices |' in md_text2

    md_text3 = ssp_writer.get_fedramp_control_tables('ac-2', 1, 1)
    assert md_text3

    md_text4 = ssp_writer.get_control_part('ac-2', 'item', 1)
    assert md_text4


def test_ssp_get_control_response(tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test generating SSP from the sample profile and generate markdown representation of it."""
    args, _, _ = setup_for_ssp(True, True, tmp_trestle_dir, prof_name, ssp_name)
    ssp_cmd = SSPGenerate()
    assert ssp_cmd._run(args) == 0
    ac1_path = tmp_trestle_dir / 'my_ssp/ac/ac-1.md'

    # set responses
    assert insert_prose(tmp_trestle_dir, 'ac-1_smt.b', 'This is a response')
    assert insert_prose(tmp_trestle_dir, 'ac-1_smt.c', 'This is also a response.')
    assert insert_prose(tmp_trestle_dir, 'ac-1_smt.a', 'This is a response.')
    assert delete_line_in_file(ac1_path, 'Add control implementation description here')
    assert delete_line_in_file(ac1_path, 'Add control implementation description here')
    assert delete_line_in_file(ac1_path, 'Add control implementation description here')

    command_ssp_assem = 'trestle author ssp-assemble -m my_ssp -o ssp_json'
    execute_command_and_assert(command_ssp_assem, 0, monkeypatch)

    ssp_json_path = tmp_trestle_dir / 'system-security-plans/ssp_json/system-security-plan.json'
    profile_path = tmp_trestle_dir / 'profiles/main_profile/profile.json'
    fetcher = cache.FetcherFactory.get_fetcher(tmp_trestle_dir, str(ssp_json_path))
    ssp_obj, _ = fetcher.get_oscal(True)

    resolved_catalog = profile_resolver.ProfileResolver.get_resolved_profile_catalog(tmp_trestle_dir, profile_path)

    ssp_io = SSPMarkdownWriter(tmp_trestle_dir)
    ssp_io.set_catalog(resolved_catalog)
    ssp_io.set_ssp(ssp_obj)

    md_text = ssp_io.get_control_response('ac-1', 1, True)
    assert md_text
    tree = MarkdownNode.build_tree_from_markdown(md_text.split('\n'))

    assert tree.get_node_for_key('## Implementation for part a.')
    assert tree.get_node_for_key('## Implementation for part c.')
    assert len(list(tree.get_all_headers_for_level(2))) == 3

    md_text = ssp_io.get_control_response('ac-1', 1, False)
    tree = MarkdownNode.build_tree_from_markdown(md_text.split('\n'))

    assert tree.get_node_for_key('## Implementation for part a.')
    assert tree.get_node_for_key('## Implementation for part c.')
    assert len(list(tree.get_all_headers_for_level(2))) == 3

    # insert some responses by component
    assert file_utils.insert_text_in_file(ac1_path, 'Implementation for part a.', '### foo comp\n')
    assert file_utils.insert_text_in_file(ac1_path, 'a response', '### bar comp\nstuff for other response\n')
    execute_command_and_assert(command_ssp_assem, 0, monkeypatch)
    ssp_obj, _ = fetcher.get_oscal(True)
    resolved_catalog = profile_resolver.ProfileResolver.get_resolved_profile_catalog(tmp_trestle_dir, profile_path)
    ssp_io = SSPMarkdownWriter(tmp_trestle_dir)
    ssp_io.set_catalog(resolved_catalog)
    ssp_io.set_ssp(ssp_obj)

    # component titles should not be there
    md_text = ssp_io.get_control_response('ac-1', 2, False, False)
    assert 'foo' not in md_text
    assert 'bar' not in md_text

    # component titles should be there
    md_text = ssp_io.get_control_response('ac-1', 2, False, True)
    assert 'foo' in md_text
    assert 'bar' in md_text


def test_writers(testdata_dir: pathlib.Path, tmp_trestle_dir: pathlib.Path, monkeypatch: MonkeyPatch) -> None:
    """Test individual writers."""
    ssp_io = SSPMarkdownWriter(tmp_trestle_dir)
    output = ssp_io._write_list_with_header('Empty body', [], 2)
    assert not output

    output = ssp_io._write_list_with_header('Sample list', ['1', '2', '3'], 2)
    tree = MarkdownNode.build_tree_from_markdown(output.split('\n'))
    node = tree.get_node_for_key('### Sample list')
    assert node
    assert node.content.text == ['', '- 1', '', '- 2', '', '- 3']

    output = ssp_io._write_table_with_header('Empty table', [[]], ['c1', 'c2', 'c3'], 0)
    assert not output

    output = ssp_io._write_table_with_header('Sample table', [['1', '2', '3'], ['4', '5', '6']], ['c1', 'c2', 'c3'], 0)
    tree = MarkdownNode.build_tree_from_markdown(output.split('\n'))
    node = tree.get_node_for_key('# Sample table')
    assert node
    assert node.content.tables[0] == '| c1 | c2 | c3 |'
    assert node.content.tables[3] == '| 4 | 5 | 6 |'

    output = ssp_io._write_str_with_header('Empty text', '', 0)
    assert not output

    output = ssp_io._write_str_with_header('Some text', 'this is a text.', 5)
    tree = MarkdownNode.build_tree_from_markdown(output.split('\n'))
    node = tree.get_node_for_key('###### Some text')
    assert node
    assert node.content.raw_text == '###### Some text\n\nthis is a text.'



# Copyright (c) 2022 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handle reading of writing controls from markdown."""
import logging
import pathlib
import re
import string
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple

import trestle.core.generic_oscal as generic
import trestle.oscal.catalog as cat
from trestle.common import const
from trestle.common.common_types import TypeWithProps
from trestle.common.err import TrestleError
from trestle.common.list_utils import as_list, delete_list_from_list, merge_dicts, none_if_empty
from trestle.common.model_utils import ModelUtils
from trestle.common.str_utils import spaces_and_caps_to_snake
from trestle.core import generators as gens
from trestle.core.control_context import ContextPurpose, ControlContext
from trestle.core.control_interface import CompDict, ComponentImpInfo, ControlInterface
from trestle.core.markdown.markdown_api import MarkdownAPI
from trestle.core.markdown.markdown_processor import MarkdownNode
from trestle.oscal import common
from trestle.oscal import component as comp
from trestle.oscal import profile as prof

logger = logging.getLogger(__name__)


class ControlReader():
    """Class to read controls from markdown."""

    @staticmethod
    def _parse_control_title_line(line: str) -> Tuple[int, str, str]:
        """Process the title line and extract the control id, group title (in brackets) and control title."""
        if line.count('-') == 0:
            raise TrestleError(f'Markdown control title format error, missing - after control id: {line}')
        split_line = line.split()
        if len(split_line) < 3 or split_line[2] != '-':
            raise TrestleError(f'Cannot parse control markdown title for control_id group and title: {line}')
        # first token after the #
        control_id = split_line[1]
        group_title_start = line.find('\[')
        group_title_end = line.find('\]')
        if group_title_start < 0 or group_title_end < 0 or group_title_start > group_title_end:
            raise TrestleError(f'unable to read group title for control {control_id}')
        group_title = line[group_title_start + 2:group_title_end].strip()
        control_title = line[group_title_end + 2:].strip()
        return control_id, group_title, control_title

    @staticmethod
    def _indent(line: str) -> int:
        """Measure indent of non-empty line."""
        if not line:
            raise TrestleError('Empty line queried for indent.')
        if line[0] not in [' ', '-']:
            return -1
        for ii in range(len(line)):
            if line[ii] == '-':
                return ii
            # if line is indented it must start with -
            if line[ii] != ' ':
                break
        raise TrestleError(f'List elements must start with -: {line}')

    @staticmethod
    def _get_next_line(ii: int, lines: List[str]) -> Tuple[int, str]:
        while ii < len(lines):
            line = lines[ii]
            if line:
                return ii, line
            ii += 1
        return -1, ''

    @staticmethod
    def _get_next_indent(ii: int, lines: List[str]) -> Tuple[int, int, str]:
        """Seek to next content line.  ii remains at line read."""
        while 0 <= ii < len(lines):
            line = lines[ii]
            if line:
                if line[0] == '#':
                    return ii, -1, line
                indent = ControlReader._indent(line)
                if indent >= 0:
                    # extract text after -
                    start = indent + 1
                    while start < len(line) and line[start] == ' ':
                        start += 1
                    if start >= len(line):
                        raise TrestleError(f'Invalid line {line}')
                    return ii, indent, line[start:]
                return ii, indent, line
            ii += 1
        return ii, -1, ''

    @staticmethod
    def _read_part_id_prose(line: str) -> Tuple[str, str]:
        """Extract the part id letter or number and prose from line."""
        start = line.find('\\[')
        end = line.find('\\]')
        prose = line.strip() if start < 0 else line[end + 2:].strip()
        id_ = '' if start < 0 or end < 0 else line[start + 2:end]
        return id_, prose

    @staticmethod
    def _bump_label(label: str) -> str:
        """
        Find next label given a string of 1 or more pure letters or digits.

        The input must be either a string of digits or a string of ascii letters - or empty string.
        """
        if not label:
            return 'a'
        if label[0] in string.digits:
            return str(int(label) + 1)
        if len(label) == 1 and label[0].lower() < 'z':
            return chr(ord(label[0]) + 1)
        # if this happens to be a string of letters, force it lowercase and bump
        label = label.lower()
        factor = 1
        value = 0
        # delta is needed because a counts as 0 when first value on right, but 1 for all others
        delta = 0
        for letter in label[::-1]:
            value += (ord(letter) - ord('a') + delta) * factor
            factor *= 26
            delta = 1

        value += 1

        new_label = ''
        delta = 0
        while value > 0:
            new_label += chr(ord('a') + value % 26 - delta)
            value = value // 26
            delta = 1
        return new_label[::-1]

    @staticmethod
    def _create_next_label(prev_label: str, indent: int) -> str:
        """
        Create new label at indent level based on previous label if available.

        If previous label is available, make this the next one in the sequence.
        Otherwise start with a or 1 on alternate levels of indentation.
        If alphabetic label reaches z, next one is aa.
        Numeric ranges from 1 to 9, then 10 etc.
        """
        if not prev_label:
            # assume indent goes in steps of 2
            return ['a', '1'][(indent // 2) % 2]
        label_prefix = ''
        label_suffix = prev_label
        is_char = prev_label[-1] in string.ascii_letters
        # if it isn't ending in letter or digit just append 'a' to end
        if not is_char and prev_label[-1] not in string.digits:
            return prev_label + 'a'
        # break in middle of string if mixed types
        if len(prev_label) > 1:
            ii = len(prev_label) - 1
            while ii >= 0:
                if prev_label[ii] not in string.ascii_letters + string.digits:
                    break
                if (prev_label[ii] in string.ascii_letters) != is_char:
                    break
                ii -= 1
            if ii >= 0:
                label_prefix = prev_label[:(ii + 1)]
                label_suffix = prev_label[(ii + 1):]

        return label_prefix + ControlReader._bump_label(label_suffix)

    @staticmethod
    def _read_parts(indent: int, ii: int, lines: List[str], parent_id: str,
                    parts: List[common.Part]) -> Tuple[int, List[common.Part]]:
        """If indentation level goes up or down, create new list or close current one."""
        while True:
            ii, new_indent, line = ControlReader._get_next_indent(ii, lines)
            if new_indent < 0:
                # we are done reading control statement
                return ii, parts
            if new_indent == indent:
                # create new item part and add to current list of parts
                id_text, prose = ControlReader._read_part_id_prose(line)
                # id_text is the part id and needs to be as a label property value
                # if none is there then create one from previous part, or use default
                if not id_text:
                    prev_label = ControlInterface.get_label(parts[-1]) if parts else ''
                    id_text = ControlReader._create_next_label(prev_label, indent)
                id_ = ControlInterface.strip_to_make_ncname(parent_id.rstrip('.') + '.' + id_text.strip('.'))
                name = 'objective' if id_.find('_obj') > 0 else 'item'
                prop = common.Property(name='label', value=id_text)
                part = common.Part(name=name, id=id_, prose=prose, props=[prop])
                parts.append(part)
                ii += 1
            elif new_indent > indent:
                # add new list of parts to last part and continue
                if len(parts) == 0:
                    raise TrestleError(f'Improper indentation structure: {line}')
                ii, new_parts = ControlReader._read_parts(new_indent, ii, lines, parts[-1].id, [])
                if new_parts:
                    parts[-1].parts = new_parts
            else:
                # return list of sub-parts
                return ii, parts

    @staticmethod
    def _read_control_statement(ii: int, lines: List[str], control_id: str) -> Tuple[int, common.Part]:
        """Search for the Control statement and read until next ## Control."""
        while 0 <= ii < len(lines) and not lines[ii].startswith(const.CONTROL_HEADER):
            ii += 1
        if ii >= len(lines):
            raise TrestleError(f'Control statement not found for control {control_id}')
        ii += 1

        ii, line = ControlReader._get_next_line(ii, lines)
        if ii < 0:
            # This means no statement and control withdrawn (this happens in NIST catalog)
            return ii, None
        if line and line[0] == ' ' and line.lstrip()[0] != '-':
            # prose that appears indented but has no - : treat it as the normal statement prose
            line = line.lstrip()
            indent = -1
            ii += 1
        else:
            ii, indent, line = ControlReader._get_next_indent(ii, lines)

        statement_id = ControlInterface.create_statement_id(control_id)
        statement_part = common.Part(name=const.STATEMENT, id=statement_id)
        # first line is either statement prose or start of statement parts
        if indent < 0:
            statement_part.prose = line
            ii += 1
        # we have absorbed possible statement prose.
        # now just read parts recursively
        # if there was no statement prose, this will re-read the line just read
        # as the start of the statement's parts
        ii, parts = ControlReader._read_parts(0, ii, lines, statement_part.id, [])
        statement_part.parts = none_if_empty(parts)
        return ii, statement_part

    @staticmethod
    def _read_control_objective(ii: int, lines: List[str], control_id: str) -> Tuple[int, Optional[common.Part]]:
        ii_orig = ii
        while 0 <= ii < len(lines) and not lines[ii].startswith(const.CONTROL_OBJECTIVE_HEADER):
            ii += 1

        if ii >= len(lines):
            return ii_orig, None
        ii += 1

        ii, line = ControlReader._get_next_line(ii, lines)
        if ii < 0:
            raise TrestleError(f'Unable to parse objective from control markdown {control_id}')
        if line and line[0] == ' ' and line.lstrip()[0] != '-':
            # prose that appears indented but has no - : treat it as the normal objective prose
            line = line.lstrip()
            indent = -1
            ii += 1
        else:
            ii, indent, line = ControlReader._get_next_indent(ii, lines)

        objective_part = common.Part(name='objective', id=f'{control_id}_obj')
        # first line is either objective prose or start of objective parts
        if indent < 0:
            objective_part.prose = line
            ii += 1
        # we have absorbed possible objective prose.
        # now just read parts recursively
        # if there was no objective prose, this will re-read the line just read
        # as the start of the objective's parts
        ii, parts = ControlReader._read_parts(0, ii, lines, objective_part.id, [])
        objective_part.parts = parts if parts else None
        return ii, objective_part

    @staticmethod
    def _read_sections(ii: int, lines: List[str], control_id: str,
                       control_parts: List[common.Part]) -> Tuple[int, Optional[List[common.Part]]]:
        """Read all sections following the section separated by ## Control."""
        new_parts = []
        prefix = const.CONTROL_HEADER + ' '
        while 0 <= ii < len(lines):
            line = lines[ii]
            if line.startswith('## What is the solution') or line.startswith(f'# {const.EDITABLE_CONTENT}'):
                ii += 1
                continue
            if not line:
                ii += 1
                continue
            if line and not line.startswith(prefix):
                # the control has no sections to read, so exit the loop
                break
            label = line[len(prefix):].strip()
            prose = ''
            ii += 1
            while 0 <= ii < len(lines) and not lines[ii].startswith(prefix) and not lines[ii].startswith(
                    f'# {const.EDITABLE_CONTENT}'):
                prose = '\n'.join([prose, lines[ii]])
                ii += 1
            if prose:
                if label.lower() == 'guidance':
                    id_ = ControlInterface.strip_to_make_ncname(control_id + '_gdn')
                else:
                    id_ = ControlInterface.strip_to_make_ncname(control_id + '_' + label)
                label = ControlInterface.strip_to_make_ncname(label)
                new_parts.append(common.Part(id=id_, name=label, prose=prose.strip('\n')))
        if new_parts:
            control_parts = [] if not control_parts else control_parts
            control_parts.extend(new_parts)
        control_parts = none_if_empty(control_parts)
        return ii, control_parts

    @staticmethod
    def _clean_prose(prose: List[str]) -> List[str]:
        # remove empty and horizontal rule lines at start and end of list of prose lines
        forward_index = 0
        for line in prose:
            if line.strip() and not line.startswith('____'):
                break
            forward_index += 1
        new_prose = prose[forward_index:]
        reverse_index = 0
        for line in reversed(new_prose):
            if line.strip() and not line.startswith('____'):
                break
            reverse_index += 1
        clean_prose = new_prose[:len(new_prose) - reverse_index]
        clean_prose = clean_prose if clean_prose else ['']
        # if there is no useful prose this will return [''] and allow generation of a statement with empty prose
        return clean_prose

    @staticmethod
    def _comp_name_in_dict(comp_name: str, comp_dict: CompDict) -> str:
        """If the name is already in the dict in a similar form, stick to that form."""
        simple_name = ControlReader.simplify_name(comp_name)
        for name in comp_dict.keys():
            if simple_name == ControlReader.simplify_name(name):
                return name
        return comp_name

    @staticmethod
    def _add_node_to_dict(
        comp_name: str,
        label: str,
        comp_dict: CompDict,
        node: MarkdownNode,
        control_id: str,
        comp_list: List[str],
        context: ControlContext
    ) -> None:
        """Extract the label, prose, possible component name - along with implementation status."""
        component_mode = context.purpose == ContextPurpose.COMPONENT
        # for ssp, ### marks component name but for component it is ##
        # if it is a header, make sure it has correct format
        if node.key and node.key[0] == '#' and ControlInterface.bad_header(node.key):
            raise TrestleError(f'Improper header format for control {control_id}: {node.key}')
        if not component_mode:
            # look for component name heading if present
            prefix = '### '
            if node.key.startswith(prefix):
                if len(node.key.split()) <= 1:
                    raise TrestleError(
                        f'Header line in control {control_id} markdown starts with {prefix} but has no content.'
                    )
                comp_name = node.key.split(' ', 1)[1].strip()
                simp_comp_name = ControlReader.simplify_name(comp_name)
                if simp_comp_name == ControlReader.simplify_name(const.SSP_MAIN_COMP_NAME) and not component_mode:
                    raise TrestleError(
                        f'Response in control {control_id} has {const.SSP_MAIN_COMP_NAME} as a component heading.  '
                        'Instead, place all response prose for the default component at the top of the section, '
                        'with no ### component_name heading.  It will be entered as prose for the default system '
                        'component.'
                    )
                if simp_comp_name in comp_list:
                    raise TrestleError(
                        f'Control {control_id} has a section with two component headings for {comp_name}.  '
                        'Please combine the sections so there is only one heading for each component in a '
                        'statement.'
                    )
                comp_list.append(simp_comp_name)
                comp_name = ControlReader._comp_name_in_dict(comp_name, comp_dict)
            elif node.key.startswith('## What is the solution'):
                comp_name = const.SSP_MAIN_COMP_NAME
                simp_comp_name = ControlReader.simplify_name(comp_name)
                comp_list.append(simp_comp_name)
                comp_name = ControlReader._comp_name_in_dict(comp_name, comp_dict)

        # prose may be empty in md and we want to capture that so put it in the comp_dict
        prose = '\n'.join(ControlReader._clean_prose(node.content.text))
        # add the prose to the comp_dict, creating new entry as needed
        if comp_name in comp_dict:
            if label in comp_dict[comp_name]:
                comp_dict[comp_name][label].prose = prose
            else:
                # create new entry with prose
                comp_dict[comp_name][label] = ComponentImpInfo(prose=prose, rules=[])
        else:
            comp_dict[comp_name] = {label: ComponentImpInfo(prose=prose, rules=[])}

        # build list of subnodes that get handled specially so they aren't processed here
        subnode_kill: List[int] = []
        status_str = None
        remarks_str = None
        for ii, subnode in enumerate(node.subnodes):
            if subnode.key.find(const.IMPLEMENTATION_STATUS_REMARKS_HEADER) >= 0:
                remarks_str = subnode.key.split(maxsplit=4)[-1]
                subnode_kill.append(ii)
            elif subnode.key.find(const.IMPLEMENTATION_STATUS_HEADER) >= 0:
                status_str = subnode.key.split(maxsplit=3)[-1]
                subnode_kill.append(ii)
            # this effectively ignores the Rules: and regards them as read-only
            elif subnode.key.find('Rules:') >= 0:
                subnode_kill.append(ii)
        if status_str:
            new_status = common.ImplementationStatus(state=status_str, remarks=remarks_str)
            if comp_name not in comp_dict:
                comp_dict[comp_name] = {}
            if label not in comp_dict[comp_name]:
                comp_dict[comp_name][label] = ComponentImpInfo(prose='', rules=[])
            comp_dict[comp_name][label].status = new_status
        delete_list_from_list(node.subnodes, subnode_kill)
        for subnode in as_list(node.subnodes):
            ControlReader._add_node_to_dict(comp_name, label, comp_dict, subnode, control_id, comp_list, context)

    @staticmethod
    def _get_statement_label(control: Optional[cat.Control], statement_id: str) -> str:
        if control:
            for part in as_list(control.parts):
                if part.name == const.STATEMENT:
                    for sub_part in as_list(part.parts):
                        if sub_part.name == 'item' and sub_part.id == statement_id:
                            return ControlInterface.get_label(sub_part)
        return ''

    @staticmethod
    def _add_component_to_dict(
        control_imp: comp.ControlImplementation, comp_title: str, control: Optional[cat.Control], comp_dict: CompDict
    ) -> Tuple[Dict[str, Dict[str, str]], List[str]]:
        """Add imp_reqs for this control and this component to the component dictionary."""
        control_id = control.id if control else 'temp'
        params_dict = {}
        all_rules = set()
        sub_comp_dict = {}
        for imp_req in ControlInterface.get_control_imp_reqs(control_imp, control_id):
            # if description is same as control id regard it as not having prose
            # add top level control guidance with no statement id
            prose = ControlReader._handle_empty_prose(imp_req.description, control_id)
            params_dict.update(ControlInterface.get_params_dict_from_item(imp_req))
            rules_list = ControlInterface.get_rule_list_for_item(imp_req)
            all_rules.update(rules_list)
            status = ControlInterface.get_status_from_props(imp_req)
            sub_comp_dict[''] = ComponentImpInfo(prose=prose, status=status, rules=rules_list)
            for statement in as_list(imp_req.statements):
                rules_list = ControlInterface.get_rule_list_for_item(statement)
                all_rules.update(rules_list)
                status = ControlInterface.get_status_from_props(statement)
                label = ControlReader._get_statement_label(control, statement.statement_id)
                prose = ControlReader._handle_empty_prose(statement.description, statement.statement_id)
                sub_comp_dict[label] = ComponentImpInfo(prose=prose, status=status, rules=rules_list)
        if sub_comp_dict:
            comp_dict[comp_title] = sub_comp_dict
        return params_dict, sorted(all_rules)

    @staticmethod
    def _insert_header_content(
        imp_req: generic.GenericImplementedRequirement, header: Dict[str, Any], control_id: str
    ) -> None:
        """Insert yaml header content into the imp_req and its by_comps."""
        dict_ = header.get(const.TRESTLE_PROPS_TAG, {})
        # if an attribute is in the dict but it is None, need to make sure we get empty list anyway
        control_orig = as_list(dict_.get(const.CONTROL_ORIGINATION, []))
        imp_status = as_list(dict_.get(const.IMPLEMENTATION_STATUS, []))
        roles = as_list(dict_.get(const.RESPONSIBLE_ROLES, []))
        props = []
        responsible_roles = []
        for co in control_orig:
            if isinstance(co, str):
                props.append(common.Property(ns=const.NAMESPACE_NIST, name=const.CONTROL_ORIGINATION, value=co))
            elif isinstance(co, dict):
                if const.STATUS_INHERITED in co:
                    uuid = co[const.STATUS_INHERITED]
                    props.append(common.Property(name=const.LEV_AUTH_UUID, value=uuid))
                    props.append(
                        common.Property(
                            ns=const.NAMESPACE_NIST, name=const.CONTROL_ORIGINATION, value=const.STATUS_INHERITED
                        )
                    )
                else:
                    raise TrestleError(f'The yaml header for control {control_id} has unexpected content: {co}')
            else:
                raise TrestleError(f'The yaml header for control {control_id} has unexpected content: {co}')
        # FIXME this needs reworking
        for status in imp_status:
            if isinstance(status, str):
                props.append(
                    common.Property(ns=const.NAMESPACE_FEDRAMP, name=const.IMPLEMENTATION_STATUS, value=status)
                )
            elif isinstance(status, dict):
                if const.STATUS_PLANNED in status:
                    if const.STATUS_COMPLETION_DATE not in status:
                        raise TrestleError(
                            f'Planned status in the control {control_id} yaml header must '
                            f'specify completion date: {status}'
                        )
                    props.append(
                        common.Property(
                            ns=const.NAMESPACE_FEDRAMP, name=const.STATUS_PLANNED, value=status[const.STATUS_PLANNED]
                        )
                    )
                    datestr = status[const.STATUS_COMPLETION_DATE]
                    datestr = datestr.strftime('%Y-%m-%d') if isinstance(datestr, datetime) else str(datestr)
                    props.append(
                        common.Property(
                            ns=const.NAMESPACE_FEDRAMP, name=const.STATUS_PLANNED_COMPLETION_DATE, value=datestr
                        )
                    )
                else:
                    if len(status) != 1:
                        raise TrestleError(f'Unexpected content in control {control_id} yaml header: {status}')
                    value = list(status.keys())[0]
                    remark = list(status.values())[0]
                    props.append(
                        common.Property(
                            ns=const.NAMESPACE_FEDRAMP,
                            name=const.IMPLEMENTATION_STATUS,
                            value=value,
                            remarks=common.Remarks(__root__=remark)
                        )
                    )
            else:
                raise TrestleError(f'Unexpected content in control {control_id} yaml header: {status}')
        for role in roles:
            if isinstance(role, str):
                # role_id must conform to NCNAME regex
                role = role.strip().replace(' ', '_')
                if role:
                    responsible_roles.append(common.ResponsibleRole(role_id=role))
            else:
                logger.warning(f'Role in header for control {control_id} not recognized: {role}')
        if props:
            imp_req.props = as_list(imp_req.props)
            imp_req.props.extend(props)
        if responsible_roles:
            imp_req.responsible_roles = as_list(imp_req.responsible_roles)
            imp_req.responsible_roles.extend(responsible_roles)
            imp_req.responsible_roles = none_if_empty(imp_req.responsible_roles)
            # enforce single list of resp. roles for control and each by_comp
            for by_comp in as_list(imp_req.by_components):
                by_comp.responsible_roles = imp_req.responsible_roles

    @staticmethod
    def simplify_name(name: str) -> str:
        """Simplify the name to ignore variations in case, space, hyphen, underscore, slash."""
        return name.lower().replace(' ', '').replace('-', '').replace('_', '').replace('/', '')

    @staticmethod
    def _get_label_from_implementation_header(imp_header: str):
        # assumed to be of form: Implementation for part a.
        split_header = imp_header.split(' ', 4)
        if len(split_header) != 5:
            raise TrestleError(f'Implementation header cannot be parsed for statement part: {imp_header}')
        return split_header[4].strip()

    @staticmethod
    def read_all_implementation_prose_and_header(
        control: cat.Control, control_file: pathlib.Path, context: ControlContext
    ) -> Tuple[CompDict, Dict[str, List[str]]]:
        """
        Find all labels and associated implementation prose in the markdown for this control.

        Args:
            control_file: path to the control markdown file
            context: context of the control usage

        Returns:
            Dictionary by comp_name of Dictionaries of part labels and corresponding prose read from the markdown file.
            Also returns the yaml header as dict in second part of tuple.
            This does not generate components - it only tracks component names and associated responses.

        Notes:
            If a component is provided, any implementation prose for a control will be added.
            In addition, the implemented requirement will be queried for a
            property corresponding to implementation status and included if available.
        """
        # this level only adds for known component but add_node_to_dict can add for other components
        comp_name = context.comp_name if context.comp_name else const.SSP_MAIN_COMP_NAME
        control_id = control_file.stem

        comp_dict: CompDict = {}
        yaml_header = {}
        # use context.rules_dict and params_dict to map rules
        if context.purpose == ContextPurpose.COMPONENT:
            # find rule info needed by this control
            params_dict, rules_list = ControlReader._add_component_to_dict(
                context.control_implementation, context.comp_name, control, comp_dict
            )
            all_params = []
            if params_dict:
                if not set(params_dict.keys()).issuperset(rules_list):
                    raise TrestleError(f'Control {control_id} has a parameter assigned to a rule that is not defined.')
                if context.rules_dict:
                    all_params.extend(
                        [
                            {
                                context.rules_dict[id_]['name']: context.rules_params_dict[id_]
                                for id_ in context.rules_params_dict.keys()
                            }
                        ]
                    )
            if context.rules_dict:
                rule_ids = [id_ for id_ in context.rules_dict.keys() if context.rules_dict[id_]['name'] in rules_list]
                yaml_header[const.COMP_DEF_RULES_TAG] = [context.rules_dict[id_] for id_ in rule_ids]
                all_params.extend(
                    [context.rules_params_dict[id_] for id_ in rule_ids if id_ in context.rules_params_dict]
                )
            if all_params:
                yaml_header[const.RULES_PARAMS_TAG] = all_params
            if context.rules_param_vals:
                yaml_header[const.COMP_DEF_RULES_PARAM_VALS_TAG] = context.rules_param_vals

        if not control_file.exists():
            return comp_dict, yaml_header
        # if the file exists, load the contents and do not use prose from comp_dict
        try:
            md_api = MarkdownAPI()
            new_yaml_header, control_md = md_api.processor.process_markdown(control_file)
            yaml_header = new_yaml_header

            # first get the header strings, including statement labels, for statement imp reqs
            imp_string = '## Implementation '
            headers = control_md.get_all_headers_for_level(2)
            # get e.g. ## Implementation a.  ## Implementation b. etc
            imp_header_list = [header for header in headers if header.startswith(imp_string)]

            # now get the (one) header for the main solution
            main_headers = list(control_md.get_all_headers_for_key(const.SSP_MD_IMPLEMENTATION_QUESTION, False))
            # should be only one header, so warn if others found
            if main_headers:
                if len(main_headers) > 1:
                    logger.warning(
                        f'Control {control_id} has {len(main_headers)} main header responses.  Will use first one only.'
                    )
                main_header = main_headers[0]
                node = control_md.get_all_nodes_for_keys([main_header], False)[0]
                # this node is top level so it will have empty label
                # it may have subnodes of Rules, Implementation Status, Implementaton Remarks
                ControlReader._add_node_to_dict(comp_name, '', comp_dict, node, control_id, [], context)
            for imp_header in imp_header_list:
                label = ControlReader._get_label_from_implementation_header(imp_header)
                node = control_md.get_node_for_key(imp_header)
                ControlReader._add_node_to_dict(comp_name, label, comp_dict, node, control_id, [], context)

        except TrestleError as e:
            raise TrestleError(f'Error occurred reading {control_file}: {e}')
        return comp_dict, yaml_header

    @staticmethod
    def _handle_empty_prose(prose: str, id_: str) -> str:
        """Regard prompt text or id_ as no prose and return blank string."""
        if prose.startswith(const.SSP_ADD_IMPLEMENTATION_PREFIX) or prose == id_:
            return ''
        return prose

    @staticmethod
    def read_implemented_requirement(
        control_file: pathlib.Path, avail_comps: Dict[str, generic.GenericComponent], context: ControlContext
    ) -> Tuple[str, generic.GenericImplementedRequirement]:
        """
        Get the implementated requirement associated with given control and link to existing components or new ones.

        Args:
            control_file: path of the control markdown file
            avail_comps: dictionary of known components keyed by component name
            context: context of the control usage

        Returns:
            Tuple: The control sort-id and the one implemented requirement for this control.

        Notes:
            Each statement may have several responses, with each response in a by_component for a specific component.
            statement_map keeps track of statements that may have several by_component responses.
        """
        control_id = control_file.stem
        comp_dict, header = ControlReader.read_all_implementation_prose_and_header(None, control_file, context)

        statement_map: Dict[str, generic.GenericStatement] = {}
        # create a new implemented requirement linked to the control id to hold the statements
        imp_req: generic.GenericImplementedRequirement = generic.GenericImplementedRequirement.generate()
        imp_req.control_id = control_id

        raw_comp_dict = {ControlReader.simplify_name(key): value for key, value in comp_dict.items()}
        raw_avail_comps = {ControlReader.simplify_name(key): value for key, value in avail_comps.items()}

        # the comp_dict captures all component names referenced by the control
        # need to create new components if not already in dict by looping over comps referenced by this control
        for comp_name in comp_dict.keys():
            component: Optional[generic.GenericComponent] = None
            raw_comp_name = ControlReader.simplify_name(comp_name)
            if raw_comp_name == ControlReader.simplify_name(const.SSP_MD_IMPLEMENTATION_QUESTION):
                comp_info: ComponentImpInfo = list(raw_comp_dict[raw_comp_name].items())[0][1]
                imp_req.description = ControlReader._handle_empty_prose(comp_info.prose, control_id)
                if comp_info.status:
                    ControlInterface.insert_status_in_props(imp_req, comp_info.status)
                continue
            if raw_comp_name in raw_avail_comps:
                component = raw_avail_comps[raw_comp_name]
            else:
                # here is where we create a new component on the fly as needed
                component = generic.GenericComponent.generate()
                component.title = comp_name
                avail_comps[comp_name] = component
                raw_avail_comps[raw_comp_name] = component
            # now create statements to hold the by-components and assign the statement id
            for label, comp_info in raw_comp_dict[raw_comp_name].items():
                # if there is a statement label create by_comp - otherwise assign status and prose to imp_req
                # create a new by-component to add to this statement
                if context.purpose == ContextPurpose.COMPONENT and not label:
                    imp_req.description = ControlReader._handle_empty_prose(comp_info.prose, control_id)
                    ControlInterface.insert_status_in_props(imp_req, comp_info.status)
                    continue
                by_comp: generic.GenericByComponent = generic.GenericByComponent.generate()
                # link it to the component uuid
                by_comp.component_uuid = component.uuid
                by_comp.implementation_status = comp_info.status
                # add the response prose to the description
                by_comp.description = ControlReader._handle_empty_prose(comp_info.prose, control_id)
                statement_id = ControlInterface.create_statement_id(control_id)
                # control level response has '' as label
                if label in ['', const.STATEMENT]:
                    statement_part_id = statement_id
                else:
                    clean_label = label.strip('.')
                    statement_part_id = ControlInterface.strip_to_make_ncname(f'{statement_id}.{clean_label}')
                if statement_part_id in statement_map:
                    statement = statement_map[statement_part_id]
                else:
                    statement: generic.GenericStatement = generic.GenericStatement.generate()
                    statement.statement_id = statement_part_id
                    statement.by_components = []
                    statement_map[statement_part_id] = statement
                statement.by_components.append(by_comp)

        imp_req.statements = list(statement_map.values())
        ControlReader._insert_header_content(imp_req, header, control_id)
        sort_id = header.get(const.SORT_ID, control_id)
        return sort_id, imp_req

    @staticmethod
    def _add_control_part(
        control_id: str,
        subnode: MarkdownNode,
        required_sections_list: List[str],
        sections_dict: Dict[str, str],
        snake_dict: Dict[str, str],
        control_parts: List[common.Part],
        found_sections: List[str],
        write_mode: bool
    ) -> bool:
        match = re.match(const.CONTROL_REGEX, subnode.key)
        if match:
            part_name_raw = match.groups(0)[0]
            prose = ControlReader._clean_prose(subnode.content.text)
            prose = '\n'.join(prose)
            # prose may be empty but make part anyway if it was in markdown
            # it also may contain sub-parts
            part_name_snake = spaces_and_caps_to_snake(part_name_raw)
            part_name = snake_dict.get(part_name_snake, part_name_snake)
            # if section is required and it hasn't been edited with prose raise error
            if not write_mode and part_name in required_sections_list and prose.startswith(
                    const.PROFILE_ADD_REQUIRED_SECTION_FOR_CONTROL_TEXT):
                missing_section = sections_dict.get(part_name, part_name)
                raise TrestleError(f'Control {control_id} is missing prose for required section {missing_section}')
            id_ = f'{control_id}_{part_name}'
            # use sections dict to find correct title otherwise leave it None
            part_title = sections_dict.get(part_name, None)
            part = common.Part(id=id_, name=part_name, prose=prose, title=part_title)
            part.parts = ControlReader._add_sub_parts(part.id, subnode)
            control_parts.append(part)
            found_sections.append(part_name)
            return True
        return False

    @staticmethod
    def _add_sub_parts(part_id: str,
                       node: MarkdownNode,
                       fixed_part_name: Optional[str] = None) -> Optional[List[common.Part]]:
        if not node.subnodes:
            return None
        parts = []
        for subnode in node.subnodes:
            # the count of hashes should be correct based on parsing already down by the markdown parser
            match = re.match(const.AFTER_HASHES_REGEX, subnode.key)
            if not match:
                raise TrestleError(f'Unexpected editable header {subnode.key} found in part {part_id}')
            part_name = match.groups(0)[0]
            part_name_snake = spaces_and_caps_to_snake(part_name)
            id_ = part_id + '.' + part_name_snake
            prose_lines = ControlReader._clean_prose(subnode.content.text)
            prose = '\n'.join(prose_lines)
            final_part_name = fixed_part_name if fixed_part_name else part_name_snake
            part = common.Part(id=id_, name=final_part_name, prose=prose)
            part.parts = ControlReader._add_sub_parts(part.id, subnode, fixed_part_name)
            parts.append(part)
        return parts

    @staticmethod
    def _add_sub_part(
        control_id: str,
        subnode: MarkdownNode,
        label_map: Dict[str, str],
        by_id_parts: Dict[str, List[common.Part]],
        sections_dict: Dict[str, str]
    ) -> None:
        """Add subnode contents to the list of by_id statement parts for the top level of the control."""
        match = re.match(const.PART_REGEX, subnode.key)
        if not match:
            raise TrestleError(f'Unexpected editable header {subnode.key} found in control {control_id}')
        by_part_label = match.groups(0)[0]
        control_label_map = label_map.get(control_id, None)
        if control_label_map is None:
            raise TrestleError(f'No label map found for control {control_id}')
        by_part_id = control_label_map.get(by_part_label, None)
        if by_part_id is None:
            raise TrestleError(f'No part id found for label {by_part_label} in control {control_id}')
        inv_map = {v: k for k, v in sections_dict.items()} if sections_dict else {}
        for node2 in as_list(subnode.subnodes):
            hash_pattern = '### '
            if node2.key.startswith(hash_pattern):
                part_name = spaces_and_caps_to_snake(node2.key.replace(hash_pattern, '', 1).strip())
                part_name = inv_map.get(part_name, part_name)
                prose = ControlReader._clean_prose(node2.content.text)
                prose = '\n'.join(prose)
                id_ = f'{by_part_id}.{part_name}'
                part = common.Part(id=id_, name=part_name, prose=prose)
                part.parts = ControlReader._add_sub_parts(part.id, node2)
            else:
                raise TrestleError(f'Unexpected header {node2.key} found in control {control_id}')
            if by_part_id not in by_id_parts:
                by_id_parts[by_part_id] = []
            by_id_parts[by_part_id].append(part)

    @staticmethod
    def _get_props_list(control_id: str, label_map: Dict[str, str],
                        yaml_header: Dict[str, Any]) -> Tuple[List[common.Property], Dict[str, List[common.Property]]]:
        """Get the list of props in the yaml header of this control as separate lists with and without by_id."""
        prop_list = yaml_header.get(const.TRESTLE_ADD_PROPS_TAG, [])
        props = []
        props_by_id = {}
        for prop_d in prop_list:
            by_id = prop_d.get('smt-part', None)
            if by_id and control_id in label_map:
                by_id = label_map[control_id].get(by_id, by_id)
            prop = common.Property(name=prop_d['name'], value=prop_d['value'], ns=prop_d.get('ns', None))
            if by_id:
                if by_id not in props_by_id:
                    props_by_id[by_id] = []
                props_by_id[by_id].append(prop)
            else:
                props.append(prop)
        return props, props_by_id

    @staticmethod
    def read_editable_content(
        control_path: pathlib.Path,
        required_sections_list: List[str],
        label_map: Dict[str, Dict[str, str]],
        sections_dict: Dict[str, str],
        write_mode: bool
    ) -> Tuple[str, List[prof.Alter], Dict[str, Any]]:
        """Get parts for the markdown control corresponding to Editable Content - along with the set-parameter dict."""
        control_id = control_path.stem
        new_alters: List[prof.Alter] = []
        snake_dict: Dict[str, str] = {}

        md_api = MarkdownAPI()
        yaml_header, control_tree = md_api.processor.process_markdown(control_path)
        # extract the sort_id if present in header
        sort_id = yaml_header.get(const.SORT_ID, control_id)
        # merge the incoming sections_dict with the one in the header, with priority to incoming
        header_sections_dict: Dict[str, str] = yaml_header.get(const.SECTIONS_TAG, {})
        merged_sections_dict = merge_dicts(header_sections_dict, sections_dict)
        # query header for mapping of short to long section names
        # create reverse lookup of long snake name to short name needed for part
        for key, value in merged_sections_dict.items():
            snake_dict[spaces_and_caps_to_snake(value)] = key
        found_sections: List[str] = []

        editable_node = None
        for header in list(control_tree.get_all_headers_for_level(1)):
            if header.startswith('# Editable'):
                editable_node = control_tree.get_node_for_key(header)
                break
        if not editable_node:
            return sort_id, [], {}

        control_parts = []
        by_id_parts = {}
        for subnode in editable_node.subnodes:
            # check if it is a part added directly to the list of parts for the control
            if not ControlReader._add_control_part(control_id,
                                                   subnode,
                                                   required_sections_list,
                                                   merged_sections_dict,
                                                   snake_dict,
                                                   control_parts,
                                                   found_sections,
                                                   write_mode):
                # otherwise add it to the list of new parts to be added to the sub-parts of a part based on by-id
                ControlReader._add_sub_part(control_id, subnode, label_map, by_id_parts, merged_sections_dict)

        missing_sections = set(required_sections_list) - set(found_sections)
        if missing_sections:
            raise TrestleError(f'Control {control_id} is missing required sections {missing_sections}')
        param_dict: Dict[str, Any] = {}
        # get set_params from the header and add to parm_dict
        header_params = yaml_header.get(const.SET_PARAMS_TAG, {})
        if header_params:
            param_dict.update(header_params)

        props, props_by_id = ControlReader._get_props_list(control_id, label_map, yaml_header)

        # When adding props without by_id it can either be starting or ending and we default to ending
        # This is the default behavior as described for implicit binding in
        # https://pages.nist.gov/OSCAL/concepts/processing/profile-resolution/
        # When adding props to a part using by_id, it is the same situation because it cannot be before or after since
        # props are not in the same list as parts

        adds: List[prof.Add] = []

        # add the parts and props at control level
        if control_parts or props:
            adds.append(prof.Add(parts=none_if_empty(control_parts), props=none_if_empty(props), position='ending'))

        # add the parts and props at the part level, by-id
        by_ids = set(by_id_parts.keys()).union(props_by_id.keys())
        for by_id in sorted(by_ids):
            parts = by_id_parts.get(by_id, None)
            props = props_by_id.get(by_id, None)
            adds.append(prof.Add(parts=parts, props=props, position='ending', by_id=by_id))

        new_alters = []
        if adds:
            new_alters = [prof.Alter(control_id=control_id, adds=adds)]
        return sort_id, new_alters, param_dict

    @staticmethod
    def _update_display_prop_namespace(item: TypeWithProps):
        """Set namespace for special property display_name."""
        for prop in as_list(item.props):
            if prop.name == const.DISPLAY_NAME:
                prop.ns = const.TRESTLE_GENERIC_NS

    @staticmethod
    def read_control(control_path: pathlib.Path, set_parameters_flag: bool) -> Tuple[cat.Control, str]:
        """Read the control and group title from the markdown file."""
        control = gens.generate_sample_model(cat.Control)
        md_api = MarkdownAPI()
        yaml_header, control_tree = md_api.processor.process_markdown(control_path)
        control_titles = list(control_tree.get_all_headers_for_level(1))
        if len(control_titles) == 0:
            raise TrestleError(f'Control markdown: {control_path} contains no control title.')

        control.id, group_title, control.title = ControlReader._parse_control_title_line(control_titles[0])

        control_headers = list(control_tree.get_all_headers_for_level(2))
        if len(control_headers) == 0:
            raise TrestleError(f'Control markdown: {control_path} contains no control statements.')

        control_statement = control_tree.get_node_for_key(control_headers[0])
        rc, statement_part = ControlReader._read_control_statement(
            0, control_statement.content.raw_text.split('\n'), control.id
        )
        if rc < 0:
            return control, group_title
        control.parts = [statement_part] if statement_part else None
        control_objective = control_tree.get_node_for_key(const.CONTROL_OBJECTIVE_HEADER)
        if control_objective is not None:
            _, objective_part = ControlReader._read_control_objective(
                0, control_objective.content.raw_text.split('\n'), control.id
            )
            if objective_part:
                if control.parts:
                    control.parts.append(objective_part)
                else:
                    control.parts = [objective_part]
        for header_key in control_tree.get_all_headers_for_key(const.CONTROL_HEADER, False):
            if header_key not in {control_headers[0], const.CONTROL_OBJECTIVE_HEADER, control_titles[0]}:
                section_node = control_tree.get_node_for_key(header_key)
                _, control.parts = ControlReader._read_sections(
                    0, section_node.content.raw_text.split('\n'), control.id, control.parts
                )
        if set_parameters_flag:
            params: Dict[str, str] = yaml_header.get(const.SET_PARAMS_TAG, [])
            if params:
                control.params = []
                for id_, param_dict in params.items():
                    param_dict['id'] = id_
                    param = ModelUtils.dict_to_parameter(param_dict)
                    # if display_name is in list of properties, set its namespace
                    ControlReader._update_display_prop_namespace(param)
                    control.params.append(param)
        if const.SORT_ID in yaml_header:
            control.props = control.props if control.props else []
            control.props.append(common.Property(name=const.SORT_ID, value=yaml_header[const.SORT_ID]))
        return control, group_title



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for trestle update action class."""

from datetime import datetime
from typing import List

from trestle.core.models.actions import UpdateAction
from trestle.core.models.elements import Element, ElementPath
from trestle.oscal import OSCAL_VERSION, common


def test_update_action(sample_nist_component_def):
    """Test update action."""
    element = Element(sample_nist_component_def)

    metadata = common.Metadata(
        **{
            'title': 'My simple catalog',
            'last-modified': datetime.now().astimezone(),
            'version': '0.0.0',
            'oscal-version': OSCAL_VERSION
        }
    )

    sub_element_path = ElementPath('component-definition.metadata')
    prev_metadata = element.get_at(sub_element_path)

    uac = UpdateAction(metadata, element, sub_element_path)

    uac.execute()

    assert element.get_at(sub_element_path) is not prev_metadata
    assert element.get_at(sub_element_path) == metadata

    uac.rollback()

    assert element.get_at(sub_element_path) == prev_metadata
    assert element.get_at(sub_element_path) is not metadata


def test_update_list_sub_element_action(sample_nist_component_def):
    """Test setting a list."""
    element = Element(sample_nist_component_def)

    parties: List[common.Party] = []
    parties.append(
        common.Party(**{
            'uuid': 'ff47836c-877c-4007-bbf3-c9d9bd805000', 'name': 'TEST1', 'type': 'organization'
        })
    )
    parties.append(
        common.Party(**{
            'uuid': 'ee88836c-877c-4007-bbf3-c9d9bd805000', 'name': 'TEST2', 'type': 'organization'
        })
    )

    sub_element_path = ElementPath('component-definition.metadata.parties.*')
    uac = UpdateAction(parties, element, sub_element_path)
    uac.execute()

    assert element.get_at(sub_element_path) == parties
